# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

import numpy as np
import random
import tensorflow as tf
import tensorflow.keras as tfk
import tensorflow.keras.backend as K
import tensorflow.keras.models as tfkm
import tensorflow.keras.optimizers as tfko
import tensorflow.keras.layers as tfkl
import tensorflow.keras.activations as tfka


class DoubleDuelingRDQN_NN(object):
    def __init__(self,
                 action_size,
                 observation_size,
                 learning_rate = 1e-5):
        self.action_size = action_size
        self.observation_size = observation_size
        self.h_size = 512

        self.lr = learning_rate
        
        self.model = None
        self.construct_q_network()

    def construct_q_network(self):
        # Defines input tensors and scalars
        self.trace_length = tf.Variable(1, dtype=tf.int32, name="trace_length")
        self.dropout_rate = tf.Variable(0.0, dtype=tf.float32, trainable=False, name="dropout_rate")
        input_mem_state = tfk.Input(dtype=tf.float32, shape=(self.h_size), name='input_mem_state')
        input_carry_state = tfk.Input(dtype=tf.float32, shape=(self.h_size), name='input_carry_state')
        input_layer = tfk.Input(dtype=tf.float32, shape=(None, self.observation_size), name='input_obs')

        # Get shapes from input_layer
        batch_size = tf.shape(input_layer)[0]
        trace_len = tf.shape(input_layer)[1]
        data_size = tf.shape(input_layer)[-1]

        # Reshape for dense processing
        input_format = tf.reshape(input_layer, (-1, input_layer.shape[-1]), name="dense_reshape")

        # Bayesian NN simulate
        lay1 = tfkl.Dropout(self.dropout_rate, name="bnn_dropout")(input_format)
        # Forward pass
        lay1 = tfkl.Dense(512, name="fc_1")(lay1)
        lay1 = tf.nn.leaky_relu(lay1, alpha=0.01, name="leak_fc_1")
        lay2 = tfkl.Dense(256, name="fc_2")(lay1)
        lay2 = tf.nn.leaky_relu(lay2, alpha=0.01, name="leak_fc_2")
        lay3 = tfkl.Dense(128, name="fc_3")(lay2)
        lay3 = tf.nn.leaky_relu(lay3, alpha=0.01, name="leak_fc_3")
        lay4 = tfkl.Dense(self.h_size, name="fc_4")(lay3)

        # Reshape to (batch_size, trace_len, data_size) for rnn
        rnn_format = tf.reshape(lay4, (batch_size, trace_len, self.h_size),name="rnn_reshape")
        # Recurring part
        lstm_layer = tfkl.LSTM(self.h_size, return_state=True, name="lstm")
        lstm_state = [input_mem_state, input_carry_state]
        lstm_output, mem_s, carry_s = lstm_layer(rnn_format, initial_state=lstm_state)

        # Advantage and Value streams
        advantage = tfkl.Dense(64, name="fc_adv")(lstm_output)
        advantage = tf.nn.leaky_relu(advantage, alpha=0.01, name="leak_adv")
        advantage = tfkl.Dense(self.action_size, name="adv")(advantage)
        advantage_mean = tf.math.reduce_mean(advantage, axis=1,
                                             keepdims=True, name="adv_mean")
        advantage = tfkl.subtract([advantage, advantage_mean], name="adv_sub")

        value = tfkl.Dense(64, name="fc_val")(lstm_output)
        value = tf.nn.leaky_relu(value, alpha=0.01, name="leak_val")
        value = tfkl.Dense(1, name="val")(value)

        Q = tf.math.add(value, advantage, name="Qout")

        # Backwards pass
        model_inputs = [input_mem_state, input_carry_state, input_layer]
        model_outputs = [Q, mem_s, carry_s]
        self.model = tfk.Model(inputs=model_inputs,
                               outputs=model_outputs,
                               name=self.__class__.__name__)
        losses = [
            self._mse_loss,
            self._no_loss,
            self._no_loss
        ]
        self.optimizer = tfko.Adam(lr=self.lr, clipnorm=1.0)
        self.model.compile(loss=losses, optimizer=self.optimizer)

    def _no_loss(self, y_true, y_pred):
        return 0.0

    def _mse_loss(self, Qnext, Q):
        loss = tf.math.reduce_mean(tf.math.square(Qnext - Q), name="loss_mse")
        return loss

    def bayesian_move(self, data, mem, carry, rate = 0.0):
        self.dropout_rate.assign(float(rate))
        self.trace_length.assign(1)

        data_input = data.reshape(1, 1, -1)
        mem_input = mem.reshape(1, -1)
        carry_input = carry.reshape(1, -1)
        model_input = [mem_input, carry_input, data_input]

        Q, mem, carry = self.model.predict(model_input, batch_size = 1)
        move = np.argmax(Q)

        return move, Q, mem, carry

    def random_move(self, data, mem, carry):
        self.trace_length.assign(1)
        self.dropout_rate.assign(0.0)

        data_input = data.reshape(1, 1, -1)
        mem_input = mem.reshape(1, -1)
        carry_input = carry.reshape(1, -1)
        model_input = [mem_input, carry_input, data_input]

        Q, mem, carry = self.model.predict(model_input, batch_size = 1) 
        move = np.random.randint(0, self.action_size)

        return move, mem, carry
        
    def predict_move(self, data, mem, carry):
        self.trace_length.assign(1)
        self.dropout_rate.assign(0.0)

        data_input = data.reshape(1, 1, -1)
        mem_input = mem.reshape(1, -1)
        carry_input = carry.reshape(1, -1)
        model_input = [mem_input, carry_input, data_input]

        Q, mem, carry = self.model.predict(model_input, batch_size = 1)
        move = np.argmax(Q)

        return move, Q, mem, carry

    def update_target_hard(self, target_model):
        this_weights = self.model.get_weights()
        target_model.set_weights(this_weights)

    def update_target_soft(self, target_model, tau=1e-2):
        tau_inv = 1.0 - tau
        # Get parameters to update
        target_params = target_model.trainable_variables
        main_params = self.model.trainable_variables

        # Update each param
        for i, var in enumerate(target_params):
            var_persist = var.value() * tau_inv
            var_update = main_params[i].value() * tau
            # Poliak averaging
            var.assign(var_update + var_persist)

    def save_network(self, path):
        # Saves model at specified path as h5 file
        # nothing has changed
        self.model.save_weights(path)
        print("Successfully saved model at: {}".format(path))

    def load_network(self, path):
        # nothing has changed
        self.model.load_weights(path)
        print("Successfully loaded network from: {}".format(path))



__all__ = [
    "DoubleDuelingDQN",
    "DoubleDuelingDQNConfig",
    "evaluate",
    "train"
]

from l2rpn_baselines.DoubleDuelingDQN.DoubleDuelingDQN import DoubleDuelingDQN
from l2rpn_baselines.DoubleDuelingDQN.DoubleDuelingDQNConfig import DoubleDuelingDQNConfig
from l2rpn_baselines.DoubleDuelingDQN.evaluate import evaluate
from l2rpn_baselines.DoubleDuelingDQN.train import train



#!/usr/bin/env python3

"""
In this file, explain how to evaluate your agent.
"""

import os
from grid2op.Runner import Runner

from l2rpn_baselines.Geirina.Geirina import Geirina
from l2rpn_baselines.utils.save_log_gif import save_log_gif

def evaluate(env,
             model_name=".",
             save_path = None,
             logs_path=None,
             nb_episode=1,
             nb_process=1,
             max_steps=-1,
             verbose=False,
             save_gif=False,
             **kwargs):
    """
    In order to submit a valid basline, it is mandatory to provide a "evaluate" function with the same signature as this one.

    Parameters
    ----------
    env: :class:`grid2op.Environment.Environment`
        The environment on which the baseline will be evaluated.

    load_path: ``str``
        The path where the model is stored. This is used by the agent when calling "agent.load)

    logs_path: ``str``
        The path where the agents results will be stored.

    nb_episode: ``int``
        Number of episodes to run for the assessment of the performance.
        By default it's 1.

    nb_process: ``int``
        Number of process to be used for the assessment of the performance.
        Should be an integer greater than 1. By defaults it's 1.

    max_steps: ``int``
        Maximum number of timestep each episode can last. It should be a positive integer or -1.
        -1 means that the entire episode is run (until the chronics is out of data or until a game over).
        By default it's -1.

    verbose: ``bool``
        verbosity of the output

    save_gif: ``bool``
        Whether or not to save a gif into each episode folder corresponding to the representation of the said episode.

    kwargs:
        Other key words arguments that you are free to use for either building the agent save it etc.

    Returns
    -------
    ``None``
    """

    TEST_DEV_ENVS = ["l2rpn_case14_sandbox", "rte_case14_realistic", "rte_case14_redisp", 
    "rte_case14_test", "case14_realistic", "case14_redisp","case14_test", "l2rpn_2019"] 
    if env.name not in TEST_DEV_ENVS:
        raise RuntimeError("This baseline only works for: {} ".format(TEST_DEV_ENVS)) 

    runner_params = env.get_params_for_runner()
    runner_params["verbose"] = verbose

    # Create the agent (this piece of code can change)
    agent = Geirina(env.action_space,
                        env.observation_space,
                        name="Gerina",
                        save_path = save_path)

    # Load weights from file (for example)
    agent.load(model_name)

    # Build runner
    runner = Runner(**runner_params,
                    agentClass=None,
                    agentInstance=agent
                    )

    # you can do stuff with your model here

    # start the runner
    res = runner.run(path_save=save_path,
                     nb_episode=nb_episode,
                     nb_process=nb_process,
                     max_iter=max_steps,
                     pbar=False)

    # Print summary
    print("Evaluation summary:")
    for _, chron_name, cum_reward, nb_time_step, max_ts in res:
        msg_tmp = "\tFor chronics located at {}\n".format(chron_name)
        msg_tmp += "\t\t - cumulative reward: {:.6f}\n".format(cum_reward)
        msg_tmp += "\t\t - number of time steps completed: {:.0f} / {:.0f}".format(nb_time_step, max_ts)
        print(msg_tmp)

    #if save_gif:
    #    save_log_gif(save_path, res)


if __name__ == "__main__":
    """
    This is a possible implementation of the eval script.
    """
    from grid2op import make
    from grid2op.Rules import AlwaysLegal

    from lightsim2grid.LightSimBackend import LightSimBackend
    backend = LightSimBackend()

    env = make("l2rpn_case14_sandbox"
               # gamerules_class=AlwaysLegal,
               #backend=backend
               )

    evaluate(env,
             model_name="geirina",
             save_path="basline_result",
             logs_path="logs-train",            
             nb_episode=1,
             nb_process=1,
             max_steps=1000,
             verbose=False,
             save_gif="basline_result")



import logging

import numpy as np
import pandas as pd

import pandapower as pp


def find_value_for_pp_bus(grid: pp.pandapowerNet,
                          voltages_from: np.ndarray,
                          voltages_to: np.ndarray,
                          bus: int):
    # map line extremity voltage readings to the correct pandapower bus
    bus_vn_kv = grid.bus.loc[bus, "vn_kv"]
    for idx, line in enumerate(grid.line.itertuples()):
        if line.from_bus == bus:
            return voltages_from[idx] / bus_vn_kv
        elif line.to_bus == bus:
            return voltages_to[idx] / bus_vn_kv
    for idx, trafo in enumerate(grid.trafo.itertuples()):
        if trafo.hv_bus == bus:
            return voltages_from[len(grid.line) + idx] / bus_vn_kv
        elif trafo.lv_bus == bus:
            return voltages_to[len(grid.line) + idx] / bus_vn_kv
    return None


def calc_losses(grid: pp.pandapowerNet):
    # calculate grid losses
    return grid.res_gen.p_mw.sum() + grid.res_ext_grid.p_mw.sum() + grid.res_sgen.p_mw.sum() - grid.load.p_mw.sum()


def make_zero_idx(elements: pd.DataFrame,
                  pd_indices,
                  offset: int = 0):
    # translate arbitrary pandas indices to 0-based numpy array indices
    return [elements.index.get_loc(lidx) + offset for lidx in pd_indices]


def run_opf(grid: pp.pandapowerNet,
            min_loss_reduction_mwt: float,
            acceptable_loading: float,
            opf_type: str,
            asset: str = "",
            logger=logging.getLogger()):
    loss_before = calc_losses(grid)
    line_loading_before = grid.res_line.loading_percent.max()
    trafo_loading_before = (grid.res_trafo.i_hv_ka / grid.trafo.max_i_ka * 100.).max()
    try:
        if opf_type == "pypower":
            pp.runopp(grid)  # pypower OPF
        elif opf_type == "powermodels":
            grid.line["in_service"] = True
            grid.trafo["in_service"] = True
            pp.runpm_ots(grid, pm_nl_solver="ipopt", pm_model="ACPPowerModel")  # PowerModels.jl OPF
            grid.line.loc[:, "in_service"] = grid.res_line.loc[:, "in_service"].values.astype(bool)
            grid.trafo.loc[:, "in_service"] = grid.res_trafo.loc[:, "in_service"].values.astype(bool)
        generation = np.array(grid.res_gen.p_mw.tolist() + grid.res_ext_grid.p_mw.tolist())
        losses_avoided = loss_before - calc_losses(grid)
        line_loading_after = grid.res_line.loading_percent.max()
        trafo_loading_after = (grid.res_trafo.i_hv_ka / grid.trafo.max_i_ka * 100.).max()
        use_opf_results = False
        # three conditions for using the OPF results:
        # 1. transformer was overloaded before and is now less overloaded
        # 2. line was overloaded before and is now less overloaded
        # 3. losses are minimized at least by "min_loss_reduction_mwt" while all loadings are acceptably low
        if trafo_loading_before > acceptable_loading and trafo_loading_after < trafo_loading_before:
            use_opf_results = True
        if line_loading_before > acceptable_loading and line_loading_after < line_loading_before:
            use_opf_results = True
        if losses_avoided > min_loss_reduction_mwt and line_loading_after < acceptable_loading and trafo_loading_after < acceptable_loading:
            use_opf_results = True
        if use_opf_results:
            logger.info(f"Losses avoided: {losses_avoided:.3f} MW (max. loading: {line_loading_after:.1f}% [{line_loading_after - line_loading_before:.1f}%] "
                        f"/ Trafo: {trafo_loading_after:.1f}% [{trafo_loading_after - trafo_loading_before:.1f}%])")
            return generation, grid.line.in_service, grid.trafo.in_service, True
        return np.zeros(len(grid.gen)), grid.line.in_service, grid.trafo.in_service, False
    except pp.optimal_powerflow.OPFNotConverged:
        asset_str = f"if asset #{asset} is out of service" if len(asset) else ""
        logger.info(f"OPF failed {asset_str}")
        return np.zeros(len(grid.gen)), grid.line.in_service, grid.trafo.in_service, False



#!/usr/bin/env python3

from l2rpn_baselines.Geirina.Geirina import Geirina

def train(env,
          name="Geirina",
          iterations=1000000,
          save_path=None,
          load_path=None
          ):
    """
    This an example function to train a baseline.

    NB： This baaseline only works for IEEE14 bus environment. For other environments, the agent should be modified accordingly.

    In order to be valid, if you chose (which is recommended) to provide a training script to help other retrain your
    baseline in different environments, or for longer period of time etc. This script should be contain the "train"
    function with at least the following arguments.

    Parameters
    ----------
    env: :class:`grid2op.Environment.Environment`
        The environmnent on which the baseline will be trained

    name: ``str``
        Fancy name you give to this baseline.

    iterations: ``int``
        Number of training iterations to perform

    save_path: ``str``
        The path where the baseline will be saved at the end of the training procedure.

    load_path: ``str``
        Path where to look for reloading the model. Use ``None`` if no model should be loaded.

    kwargs:
        Other key-word arguments that you might use for training.

    """
    TEST_DEV_ENVS = ["l2rpn_case14_sandbox", "rte_case14_realistic", "rte_case14_redisp", 
    "rte_case14_test", "case14_realistic", "case14_redisp","case14_test", "l2rpn_2019"] 
    if env.name not in TEST_DEV_ENVS:
        raise RuntimeError("This baseline only works for: {} ".format(TEST_DEV_ENVS)) 

    baseline = Geirina(env.action_space,
                        env.observation_space,
                        name=name,
                        save_path=save_path)

    if load_path is not None:
        baseline.load(load_path)

    baseline.train(env, iterations)
    # as in our example (and in our explanation) we recommend to save the mode regurlarly in the "train" function
    # it is not necessary to save it again here. But if you chose not to follow these advice, it is more than
    # recommended to save the "baseline" at the end of this function with:
    # baseline.save(path_save)


if __name__ == "__main__":
    """
    This is a possible implementation of the train script.
    """
    from grid2op import make
    from grid2op.Rules import AlwaysLegal

    from lightsim2grid.LightSimBackend import LightSimBackend
    backend = LightSimBackend()

    env = make("l2rpn_case14_sandbox"
               # gamerules_class=AlwaysLegal,
               #backend=backend
               )

    train(env,
          name="Geirina",
          iterations=1000000, # this is apparently not used !
          save_path="basline_result")



# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.
import os
import json
from l2rpn_baselines.utils.BaseDeepQ import BaseDeepQ


class NNParam(object):
    """
    This class provides an easy way to save and restore, as json, the shape of your neural networks
    (number of layers, non linearities, size of each layers etc.)

    It is recommended to overload this class for each specific model.

    Attributes
    ----------

    nn_class: :class:`l2rpn_baselines.BaseDeepQ`
        The neural network class that will be created with each call of :func:`l2rpn_baselines.make_nn`

    observation_size: ``int``
        The size of the observation space.

    action_size: ``int``
        The size of the action space.

    sizes: ``list``
        A list of integer, each will represent the number of hidden units. The number of hidden layer is given by
        the size / length of this list.

    activs: ``list``
        List of activation functions (given as string). It should have the same length as the :attr:`NNParam.sizes`.
        This function should be name of keras activation function.

    list_attr_obs: ``list``
        List of the attributes that will be used from the observation and concatenated to be fed to the neural network.

    """

    _int_attr = ["action_size", "observation_size"]
    _float_attr = []
    _str_attr = []
    _list_float = []
    _list_str = ["activs", "list_attr_obs"]
    _list_int = ["sizes"]
    nn_class = BaseDeepQ

    def __init__(self,
                 action_size,
                 observation_size,
                 sizes,
                 activs,
                 list_attr_obs,
                 ):
        self.observation_size = observation_size
        self.action_size = action_size
        self.sizes = [int(el) for el in sizes]
        self.activs = [str(el) for el in activs]
        if len(self.sizes) != len(self.activs):
            raise RuntimeError("\"sizes\" and \"activs\" lists have not the same size. It's not clear how many layers "
                               "you want your neural network to have.")
        self.list_attr_obs = [str(el) for el in list_attr_obs]

    @classmethod
    def get_path_model(cls, path, name=None):
        """get the path at which the model will be saved"""
        return cls.nn_class.get_path_model(path, name=name)

    def make_nn(self, training_param):
        """build the appropriate BaseDeepQ"""
        res = self.nn_class(self, training_param)
        return res

    @staticmethod
    def get_obs_size(env, list_attr_name):
        """get the size of the flatten observation"""
        res = 0
        for obs_attr_name in list_attr_name:
            beg_, end_, dtype_ = env.observation_space.get_indx_extract(obs_attr_name)
            res += end_ - beg_  # no "+1" needed because "end_" is exclude by python convention
        return res

    def get_obs_attr(self):
        """get the names of the observation attributes that will be extracted """
        return self.list_attr_obs

    # utilitaries, do not change
    def to_dict(self):
        """convert this instance to a dictionnary"""
        # TODO copy and paste from TrainingParam
        res = {}
        for attr_nm in self._int_attr:
            tmp = getattr(self, attr_nm)
            if tmp is not None:
                res[attr_nm] = int(tmp)
            else:
                res[attr_nm] = None
        for attr_nm in self._float_attr:
            tmp = getattr(self, attr_nm)
            if tmp is not None:
                res[attr_nm] = float(tmp)
            else:
                res[attr_nm] = None
        for attr_nm in self._str_attr:
            tmp = getattr(self, attr_nm)
            if tmp is not None:
                res[attr_nm] = str(tmp)
            else:
                res[attr_nm] = None

        for attr_nm in self._list_float:
            tmp = getattr(self, attr_nm)
            res[attr_nm] = [float(el) for el in tmp]
        for attr_nm in self._list_int:
            tmp = getattr(self, attr_nm)
            res[attr_nm] = [int(el) for el in tmp]
        for attr_nm in self._list_str:
            tmp = getattr(self, attr_nm)
            res[attr_nm] = [str(el) for el in tmp]
        return res

    @classmethod
    def from_dict(cls, tmp):
        """load from a dictionnary"""
        # TODO copy and paste from TrainingParam (more or less)
        cls_as_dict = {}
        for attr_nm in cls._int_attr:
            if attr_nm in tmp:
                tmp_ = tmp[attr_nm]
                if tmp_ is not None:
                    cls_as_dict[attr_nm] = int(tmp_)
                else:
                    cls_as_dict[attr_nm] = None

        for attr_nm in cls._float_attr:
            if attr_nm in tmp:
                tmp_ = tmp[attr_nm]
                if tmp_ is not None:
                    cls_as_dict[attr_nm] = float(tmp_)
                else:
                    cls_as_dict[attr_nm] = None

        for attr_nm in cls._str_attr:
            if attr_nm in tmp:
                tmp_ = tmp[attr_nm]
                if tmp_ is not None:
                    cls_as_dict[attr_nm] = str(tmp_)
                else:
                    cls_as_dict[attr_nm] = None

        for attr_nm in cls._list_float:
            if attr_nm in tmp:
                cls_as_dict[attr_nm] = [float(el) for el in tmp[attr_nm]]
        for attr_nm in cls._list_int:
            if attr_nm in tmp:
                cls_as_dict[attr_nm] = [int(el) for el in tmp[attr_nm]]
        for attr_nm in cls._list_str:
            if attr_nm in tmp:
                cls_as_dict[attr_nm] = [str(el) for el in tmp[attr_nm]]

        res = cls(**cls_as_dict)
        return res

    @classmethod
    def from_json(cls, json_path):
        """load from a json file"""
        # TODO copy and paste from TrainingParam
        if not os.path.exists(json_path):
            raise FileNotFoundError("No path are located at \"{}\"".format(json_path))
        with open(json_path, "r") as f:
            dict_ = json.load(f)
        return cls.from_dict(dict_)

    def save_as_json(self, path, name=None):
        """save as a json file"""
        # TODO copy and paste from TrainingParam
        res = self.to_dict()
        if name is None:
            name = "neural_net_parameters.json"
        if not os.path.exists(path):
            raise RuntimeError("Directory \"{}\" not found to save the NN parameters".format(path))
        if not os.path.isdir(path):
            raise NotADirectoryError("\"{}\" should be a directory".format(path))
        path_out = os.path.join(path, name)
        with open(path_out, "w", encoding="utf-8") as f:
            json.dump(res, fp=f, indent=4, sort_keys=True)



# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

from grid2op.Agent import DoNothingAgent


class Template(DoNothingAgent):
    """
    Note that a Baseline should always somehow inherit from :class:`grid2op.Agent.BaseAgent`.

    It serves as a template agent to explain how a baseline can be built.

    As opposed to bare grid2op Agent, baselines have 3 more methods:
    - :func:`Template.load`: to load the agent, if applicable
    - :func:`Template.save`: to save the agent, if applicable
    - :func:`Template.train`: to train the agent, if applicable

    The method :func:`Template.reset` is already present in grid2op but is emphasized here. It is called
    by a runner at the beginning of each episode with the first observation.

    The method :func:`Template.act` is also present in grid2op, of course. It the main method of the baseline,
    that receives an observation (and a reward and flag that says if an episode is over or not) an return a valid
    action.

    **NB** the "real" instance of environment on which the baseline will be evaluated will be built AFTER the creation
    of the baseline. The parameters of the real environment on which the baseline will be assessed will belong to the
    same class than the argument used by the baseline. This means that if a baseline is built with a grid2op
    environment "env", this environment will not be modified in any manner, all it's internal variable will not
    change etc. This is done to prevent cheating.

    """
    def __init__(self,
                 action_space,
                 observation_space,
                 name,
                 **kwargs):
        DoNothingAgent.__init__(self, action_space)
        self.do_nothing = self.action_space()
        self.name = name

    def act(self, observation, reward, done):
        """
        This is the main method of an Template. Given the current observation and the current reward
        (ie the reward that the environment send to the agent after the previous action has been implemented).

        Parameters
        ----------
        observation: :class:`grid2op.Observation.Observation`
            The current observation of the :class:`grid2op.Environment.Environment`

        reward: ``float``
            The current reward. This is the reward obtained by the previous action

        done: ``bool``
            Whether the episode has ended or not. Used to maintain gym compatibility

        Returns
        -------
        res: :class:`grid2op.Action.PlayableAction`
            The action chosen by the bot / controler / agent.

        """
        return self.do_nothing

    def reset(self, observation):
        """
        This method is called at the beginning of a new episode.
        It is implemented by baselines to reset their internal state if needed.

        Attributes
        -----------
        obs: :class:`grid2op.Observation.BaseObservation`
            The first observation corresponding to the initial state of the environment.
        """
        pass

    def load(self, path):
        """
        This function is used to build a baseline from a folder for example. It is recommended that this load
        function give different resulting depending on the :attr:`Template.name` of the baseline.
        For example, weights of a neural network can be saved under different names that ... depends on the
        name of the instance.

        If path is ``None`` is should be undertood as "don't load anything".

        Parameters
        ----------
        path: ``str``
            the path from which load the baseline.
        """
        pass

    def save(self, path):
        """
        This method is used to store the internal state of the baseline.

        Parameters
        ----------
        path: ``str``
            The location were to store the data of the baseline. If ``None`` it should be understood as "don't save".
            In any other cases it is more than recommended that, if "baseline" is a
            baseline, then:

            .. code-block:: python3

                path = "."  # or any other
                baseline.load(path)
                loaded_baseline = Template(...)  # built with the same parameters as "baseline"
                loaded_baseline.load(path)

            is a perfectly valid script (**eg** it will work perfectly) and that after loading, any call to
            "loaded_baseline.act" will give the results as the original "baseline.act". Or in other words, "baseline"
            and "loaded_baseline" represent the same Baseline, even though they are different instances of Baseline.
        """
        pass

    def train(self, env,
              iterations,
              save_path,
              **kwargs):
        """
        This function, if provided is used to train the baseline. Make sure to save it regularly with "baseline.save"
        for example.

        At the end of the training, it is r

        Parameters
        ----------
        env: :class:`grid2op.Environment.Environment`
            The environment used to train your baseline.

        iterations: ``int``
            Number of training iterations used to train the baseline.

        save_path: ``str``
            Path were the final version of the baseline (**ie** after the "num_training_steps" training steps will
            be performed). It is more than recommended to save the results regurlarly during training, and to save
            the baseline at this location at the end.

        kwargs:
            Other key-words arguments used for training.

        Returns
        -------

        """
        # do the training as you want
        pass
        # don't forget to save your agent at the end!



# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

import os
import json
import math
import numpy as np
import tensorflow as tf
from grid2op.Agent import AgentWithConverter
from grid2op.Converter import IdToAct

from l2rpn_baselines.DoubleDuelingDQN.DoubleDuelingDQNConfig import DoubleDuelingDQNConfig as cfg
from l2rpn_baselines.DoubleDuelingDQN.DoubleDuelingDQN_NN import DoubleDuelingDQN_NN
from l2rpn_baselines.DoubleDuelingDQN.prioritized_replay_buffer import PrioritizedReplayBuffer

class DoubleDuelingDQN(AgentWithConverter):
    def __init__(self,
                 observation_space,
                 action_space,
                 name=__name__,
                 is_training=False):
        # Call parent constructor
        AgentWithConverter.__init__(self, action_space,
                                    action_space_converter=IdToAct)
        self.obs_space = observation_space

        # Filter
        #print("Actions filtering...")
        self.action_space.filter_action(self._filter_action)
        #print("..Done")
        
        # Store constructor params
        self.name = name
        self.num_frames = cfg.N_FRAMES
        self.is_training = is_training
        self.batch_size = cfg.BATCH_SIZE
        self.lr = cfg.LR
        
        # Declare required vars
        self.Qmain = None
        self.obs = None
        self.state = []
        self.frames = []

        # Declare training vars
        self.per_buffer = None
        self.done = False
        self.frames2 = None
        self.epoch_rewards = None
        self.epoch_alive = None
        self.Qtarget = None
        self.epsilon = 0.0

        # Compute dimensions from intial spaces
        self.observation_size = self.obs_space.size_obs()
        self.action_size = self.action_space.size()

        # Load network graph
        self.Qmain = DoubleDuelingDQN_NN(self.action_size,
                                         self.observation_size,
                                         num_frames=self.num_frames,
                                         learning_rate=self.lr,
                                         learning_rate_decay_steps=cfg.LR_DECAY_STEPS,
                                         learning_rate_decay_rate=cfg.LR_DECAY_RATE)
        # Setup training vars if needed
        if self.is_training:
            self._init_training()

    def _filter_action(self, action):
        MAX_ELEM = 2
        act_dict = action.impact_on_objects()
        elem = 0
        elem += act_dict["force_line"]["reconnections"]["count"]
        elem += act_dict["force_line"]["disconnections"]["count"]
        elem += act_dict["switch_line"]["count"]
        elem += len(act_dict["topology"]["bus_switch"])
        elem += len(act_dict["topology"]["assigned_bus"])
        elem += len(act_dict["topology"]["disconnect_bus"])
        elem += len(act_dict["redispatch"]["generators"])

        if elem <= MAX_ELEM:
            return True
        return False
            
    def _init_training(self):
        self.epsilon = cfg.INITIAL_EPSILON
        self.frames2 = []
        self.epoch_rewards = []
        self.epoch_alive = []
        self.per_buffer = PrioritizedReplayBuffer(cfg.PER_CAPACITY, cfg.PER_ALPHA)
        self.Qtarget = DoubleDuelingDQN_NN(self.action_size,
                                           self.observation_size,
                                           num_frames = self.num_frames)

    def _reset_state(self, current_obs):
        # Initial state
        self.obs = current_obs
        self.state = self.convert_obs(self.obs)
        self.done = False

    def _reset_frame_buffer(self):
        # Reset frame buffers
        self.frames = []
        if self.is_training:
            self.frames2 = []

    def _save_current_frame(self, state):
        self.frames.append(state.copy())
        if len(self.frames) > self.num_frames:
            self.frames.pop(0)

    def _save_next_frame(self, next_state):
        self.frames2.append(next_state.copy())
        if len(self.frames2) > self.num_frames:
            self.frames2.pop(0)

    def _adaptive_epsilon_decay(self, step):
        ada_div = cfg.DECAY_EPSILON / 10.0
        step_off = step + ada_div
        ada_eps = cfg.INITIAL_EPSILON * -math.log10((step_off + 1) / (cfg.DECAY_EPSILON + ada_div))
        ada_eps_up_clip = min(cfg.INITIAL_EPSILON, ada_eps)
        ada_eps_low_clip = max(cfg.FINAL_EPSILON, ada_eps_up_clip)
        return ada_eps_low_clip
            
    def _save_hyperparameters(self, logpath, env, steps):
        r_instance = env.reward_helper.template_reward
        hp = {
            "lr": cfg.LR,
            "lr_decay_steps": cfg.LR_DECAY_STEPS,
            "lr_decay_rate": cfg.LR_DECAY_RATE,
            "batch_size": cfg.BATCH_SIZE,
            "stack_frames": cfg.N_FRAMES,
            "iter": steps,
            "e_start": cfg.INITIAL_EPSILON,
            "e_end": cfg.FINAL_EPSILON,
            "e_decay": cfg.DECAY_EPSILON,
            "discount": cfg.DISCOUNT_FACTOR,
            "per_alpha": cfg.PER_ALPHA,
            "per_beta": cfg.PER_BETA,
            "per_capacity": cfg.PER_CAPACITY,
            "update_freq": cfg.UPDATE_FREQ,
            "update_hard": cfg.UPDATE_TARGET_HARD_FREQ,
            "update_soft": cfg.UPDATE_TARGET_SOFT_TAU,
            "reward": dict(r_instance)
        }
        hp_filename = "{}-hypers.json".format(self.name)
        hp_path = os.path.join(logpath, hp_filename)
        with open(hp_path, 'w') as fp:
            json.dump(hp, fp=fp, indent=2)

    ## Agent Interface
    def convert_obs(self, observation):
        li_vect=  []
        for el in observation.attr_list_vect:
            v = observation._get_array_from_attr_name(el).astype(np.float32)
            v_fix = np.nan_to_num(v)
            v_norm = np.linalg.norm(v_fix)
            if v_norm > 1e6:
                v_res = (v_fix / v_norm) * 10.0
            else:
                v_res = v_fix
            li_vect.append(v_res)
        return np.concatenate(li_vect)

    def convert_act(self, action):
        return super().convert_act(action)

    ## Baseline Interface
    def reset(self, observation):
        self._reset_state(observation)
        self._reset_frame_buffer()

    def my_act(self, state, reward, done=False):
        # Register current state to stacking buffer
        self._save_current_frame(state)
        # We need at least num frames to predict
        if len(self.frames) < self.num_frames:
            return 0 # Do nothing
        # Infer with the last num_frames states
        a, _ = self.Qmain.predict_move(np.array(self.frames))
        return a
    
    def load(self, path):
        self.Qmain.load_network(path)
        if self.is_training:
            self.Qmain.update_target_hard(self.Qtarget.model)

    def save(self, path):
        self.Qmain.save_network(path)

    ## Training Procedure
    def train(self, env,
              iterations,
              save_path,
              num_pre_training_steps=0,
              logdir = "logs-train"):
        # Make sure we can fill the experience buffer
        if num_pre_training_steps < self.batch_size * self.num_frames:
            num_pre_training_steps = self.batch_size * self.num_frames

        # Loop vars
        num_training_steps = iterations
        num_steps = num_pre_training_steps + num_training_steps
        step = 0
        self.epsilon = cfg.INITIAL_EPSILON
        alive_steps = 0
        total_reward = 0
        self.done = True

        # Create file system related vars
        logpath = os.path.join(logdir, self.name)
        os.makedirs(save_path, exist_ok=True)
        modelpath = os.path.join(save_path, self.name + ".h5")
        self.tf_writer = tf.summary.create_file_writer(logpath, name=self.name)
        self._save_hyperparameters(save_path, env, num_steps)

        # Training loop
        while step < num_steps:
            # Init first time or new episode
            if self.done:
                new_obs = env.reset() # This shouldn't raise
                self.reset(new_obs)
            if cfg.VERBOSE and step % 1000 == 0:
                print("Step [{}] -- Random [{}]".format(step, self.epsilon))

            # Save current observation to stacking buffer
            self._save_current_frame(self.state)

            # Choose an action
            if step <= num_pre_training_steps:
                a = self.Qmain.random_move()
            elif np.random.rand(1) < self.epsilon:
                a = self.Qmain.random_move()
            elif len(self.frames) < self.num_frames:
                a = 0 # Do nothing
            else:
                a, _ = self.Qmain.predict_move(np.array(self.frames))

            # Convert it to a valid action
            act = self.convert_act(a)
            # Execute action
            new_obs, reward, self.done, info = env.step(act)
            new_state = self.convert_obs(new_obs)
            if info["is_illegal"] or info["is_ambiguous"] or \
               info["is_dispatching_illegal"] or info["is_illegal_reco"]:
                if cfg.VERBOSE:
                    print (a, info)

            # Save new observation to stacking buffer
            self._save_next_frame(new_state)

            # Save to experience buffer
            if len(self.frames2) == self.num_frames:
                self.per_buffer.add(np.array(self.frames),
                                    a, reward,
                                    np.array(self.frames2),
                                    self.done)

            # Perform training when we have enough experience in buffer
            if step >= num_pre_training_steps:
                training_step = step - num_pre_training_steps
                # Decay chance of random action
                self.epsilon = self._adaptive_epsilon_decay(training_step)

                # Perform training at given frequency
                if step % cfg.UPDATE_FREQ == 0 and \
                   len(self.per_buffer) >= self.batch_size:
                    # Perform training
                    self._batch_train(training_step, step)

                    if cfg.UPDATE_TARGET_SOFT_TAU > 0.0:
                        tau = cfg.UPDATE_TARGET_SOFT_TAU
                        # Update target network towards primary network
                        self.Qmain.update_target_soft(self.Qtarget.model, tau)

                # Every UPDATE_TARGET_HARD_FREQ trainings, update target completely
                if cfg.UPDATE_TARGET_HARD_FREQ > 0 and \
                   step % (cfg.UPDATE_FREQ * cfg.UPDATE_TARGET_HARD_FREQ) == 0:
                    self.Qmain.update_target_hard(self.Qtarget.model)

            total_reward += reward
            if self.done:
                self.epoch_rewards.append(total_reward)
                self.epoch_alive.append(alive_steps)
                if cfg.VERBOSE:
                    print("Survived [{}] steps".format(alive_steps))
                    print("Total reward [{}]".format(total_reward))
                alive_steps = 0
                total_reward = 0
            else:
                alive_steps += 1
            
            # Save the network every 1000 iterations
            if step > 0 and step % 1000 == 0:
                self.save(modelpath)

            # Iterate to next loop
            step += 1
            # Make new obs the current obs
            self.obs = new_obs
            self.state = new_state

        # Save model after all steps
        self.save(modelpath)

    def _batch_train(self, training_step, step):
        """Trains network to fit given parameters"""

        # Sample from experience buffer
        sample_batch = self.per_buffer.sample(self.batch_size, cfg.PER_BETA)
        s_batch = sample_batch[0]
        a_batch = sample_batch[1]
        r_batch = sample_batch[2]
        s2_batch = sample_batch[3]
        d_batch = sample_batch[4]
        w_batch = sample_batch[5]
        idx_batch = sample_batch[6]

        Q = np.zeros((self.batch_size, self.action_size))

        # Reshape frames to 1D
        input_size = self.observation_size * self.num_frames
        input_t = np.reshape(s_batch, (self.batch_size, input_size))
        input_t_1 = np.reshape(s2_batch, (self.batch_size, input_size))

        # Save the graph just the first time
        if training_step == 0:
            tf.summary.trace_on()

        # T Batch predict
        Q = self.Qmain.model.predict(input_t, batch_size = self.batch_size)

        ## Log graph once and disable graph logging
        if training_step == 0:
            with self.tf_writer.as_default():
                tf.summary.trace_export(self.name + "-graph", step)

        # T+1 batch predict
        Q1 = self.Qmain.model.predict(input_t_1, batch_size=self.batch_size)
        Q2 = self.Qtarget.model.predict(input_t_1, batch_size=self.batch_size)

        # Compute batch Qtarget using Double DQN
        for i in range(self.batch_size):
            doubleQ = Q2[i, np.argmax(Q1[i])]
            Q[i, a_batch[i]] = r_batch[i]
            if d_batch[i] == False:
                Q[i, a_batch[i]] += cfg.DISCOUNT_FACTOR * doubleQ

        # Batch train
        loss = self.Qmain.train_on_batch(input_t, Q, w_batch)

        # Update PER buffer
        priorities = self.Qmain.batch_sq_error
        # Can't be zero, no upper limit
        priorities = np.clip(priorities, a_min=1e-8, a_max=None)
        self.per_buffer.update_priorities(idx_batch, priorities)

        # Log some useful metrics every even updates
        if step % (cfg.UPDATE_FREQ * 2) == 0:
            with self.tf_writer.as_default():
                mean_reward = np.mean(self.epoch_rewards)
                mean_alive = np.mean(self.epoch_alive)
                if len(self.epoch_rewards) >= 100:
                    mean_reward_100 = np.mean(self.epoch_rewards[-100:])
                    mean_alive_100 = np.mean(self.epoch_alive[-100:])
                else:
                    mean_reward_100 = mean_reward
                    mean_alive_100 = mean_alive
                tf.summary.scalar("mean_reward", mean_reward, step)
                tf.summary.scalar("mean_alive", mean_alive, step)
                tf.summary.scalar("mean_reward_100", mean_reward_100, step)
                tf.summary.scalar("mean_alive_100", mean_alive_100, step)
                tf.summary.scalar("loss", loss, step)
                tf.summary.scalar("lr", self.Qmain.train_lr, step)
            if cfg.VERBOSE:
                print("loss =", loss)



#!/usr/bin/env python3

# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

import os
import argparse
import tensorflow as tf

from grid2op.MakeEnv import make
from grid2op.Runner import Runner
from grid2op.Reward import *
from grid2op.Action import *

from l2rpn_baselines.SliceRDQN.SliceRDQN import SliceRDQN as RDQNAgent
from l2rpn_baselines.utils.save_log_gif import save_log_gif

DEFAULT_LOGS_DIR = "./logs-eval"
DEFAULT_NB_EPISODE = 1
DEFAULT_NB_PROCESS = 1
DEFAULT_MAX_STEPS = -1
DEFAULT_VERBOSE = True

def cli():
    parser = argparse.ArgumentParser(description="Eval baseline DDDQN")
    parser.add_argument("--data_dir", required=True,
                        help="Path to the dataset root directory")
    parser.add_argument("--load_file", required=True,
                        help="The path to the model [.h5]")
    parser.add_argument("--logs_dir", required=False,
                        default=DEFAULT_LOGS_DIR, type=str,
                        help="Path to output logs directory")
    parser.add_argument("--nb_episode", required=False,
                        default=DEFAULT_NB_EPISODE, type=int,
                        help="Number of episodes to evaluate")
    parser.add_argument("--nb_process", required=False,
                        default=DEFAULT_NB_PROCESS, type=int,
                        help="Number of cores to use")
    parser.add_argument("--max_steps", required=False,
                        default=DEFAULT_MAX_STEPS, type=int,
                        help="Maximum number of steps per scenario")
    parser.add_argument("--gif", action='store_true',
                        help="Enable GIF Output")
    parser.add_argument("--verbose", action='store_true',
                        help="Verbose runner output")
    return parser.parse_args()


def evaluate(env,
             load_path=None,
             logs_path=DEFAULT_LOGS_DIR,
             nb_episode=DEFAULT_NB_EPISODE,
             nb_process=DEFAULT_NB_PROCESS,
             max_steps=DEFAULT_MAX_STEPS,
             verbose=DEFAULT_VERBOSE,
             save_gif=False):

    # Limit gpu usage
    physical_devices = tf.config.list_physical_devices('GPU')
    if len(physical_devices):
        tf.config.experimental.set_memory_growth(physical_devices[0], True)

    runner_params = env.get_params_for_runner()
    runner_params["verbose"] = verbose

    # Run
    # Create agent
    agent = RDQNAgent(env.observation_space,
                      env.action_space,
                      is_training=False)

    # Load weights from file
    agent.load(load_path)

    # Build runner
    runner = Runner(**runner_params,
                    agentClass=None,
                    agentInstance=agent)

    # Print model summary
    if verbose:
        stringlist = []
        agent.Qmain.model.summary(print_fn=lambda x: stringlist.append(x))
        short_model_summary = "\n".join(stringlist)
        print(short_model_summary)

    # Run
    os.makedirs(logs_path, exist_ok=True)
    res = runner.run(path_save=logs_path,
                     nb_episode=nb_episode,
                     nb_process=nb_process,
                     max_iter=max_steps,
                     pbar=verbose)

    # Print summary
    if verbose:
        print("Evaluation summary:")
        for _, chron_name, cum_reward, nb_time_step, max_ts in res:
            msg_tmp = "chronics at: {}".format(chron_name)
            msg_tmp += "\ttotal reward: {:.6f}".format(cum_reward)
            msg_tmp += "\ttime steps: {:.0f}/{:.0f}".format(nb_time_step,
                                                            max_ts)
            print(msg_tmp)

    if save_gif:
        save_log_gif(logs_path, res)

    return res


if __name__ == "__main__":
    # Parse command line
    args = cli()
    # Create dataset env
    env = make(args.data_dir,
               reward_class=RedispReward,
               action_class=TopologyAction,
               other_rewards={
                   "bridge": BridgeReward,
                   "overflow": CloseToOverflowReward,
                   "distance": DistanceReward
               })
    # Call evaluation interface
    evaluate(env,
             load_path=args.load_file,
             logs_path=args.logs_dir,
             nb_episode=args.nb_episode,
             nb_process=args.nb_process,
             max_steps=args.max_steps,
             verbose=args.verbose,
             save_gif=args.gif)



#!/usr/bin/env python3

# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

import os
import warnings
import tensorflow as tf

from l2rpn_baselines.utils import cli_train
from l2rpn_baselines.DuelQSimple.DuelQSimple import DuelQSimple, DEFAULT_NAME
from l2rpn_baselines.DuelQSimple.DuelQ_NNParam import DuelQ_NNParam
from l2rpn_baselines.DuelQSimple.DuelQ_NN import DuelQ_NN
from l2rpn_baselines.utils import TrainingParam
from l2rpn_baselines.utils.waring_msgs import _WARN_GPU_MEMORY


def train(env,
          name=DEFAULT_NAME,
          iterations=1,
          save_path=None,
          load_path=None,
          logs_dir=None,
          training_param=None,
          filter_action_fun=None,
          verbose=True,
          kwargs_converters={},
          kwargs_archi={}):

    """
    This function implements the "training" part of the balines "DuelQSimple".

    Parameters
    ----------
    env: :class:`grid2op.Environment`
        Then environment on which you need to train your agent.

    name: ``str```
        The name of your agent.

    iterations: ``int``
        For how many iterations (steps) do you want to train your agent. NB these are not episode, these are steps.

    save_path: ``str``
        Where do you want to save your baseline.

    load_path: ``str``
        If you want to reload your baseline, specify the path where it is located. **NB** if a baseline is reloaded
        some of the argument provided to this function will not be used.

    logs_dir: ``str``
        Where to store the tensorboard generated logs during the training. ``None`` if you don't want to log them.

    verbose: ``bool``
        If you want something to be printed on the terminal (a better logging strategy will be put at some point)

    training_param: :class:`l2rpn_baselines.utils.TrainingParam`
        The parameters describing the way you will train your model.

    filter_action_fun: ``function``
        A function to filter the action space. See
        `IdToAct.filter_action <https://grid2op.readthedocs.io/en/latest/converter.html#grid2op.Converter.IdToAct.filter_action>`_
        documentation.

    kwargs_converters: ``dict``
        A dictionary containing the key-word arguments pass at this initialization of the
        :class:`grid2op.Converter.IdToAct` that serves as "Base" for the Agent.

    kwargs_archi: ``dict``
        Key word arguments used for making the :class:`DeepQ_NNParam` object that will be used to build the baseline.

    Returns
    -------

    baseline: :class:`DeepQSimple`
        The trained baseline.


    .. _Example-duelqsimple:

    Examples
    ---------

    Here is an example on how to train a DuelQSimple baseline.

    First define a python script, for example

    .. code-block:: python

        import grid2op
        from grid2op.Reward import L2RPNReward
        from l2rpn_baselines.utils import TrainingParam, NNParam
        from l2rpn_baselines.DuelQSimple import train

        # define the environment
        env = grid2op.make("l2rpn_case14_sandbox",
                           reward_class=L2RPNReward)

        # use the default training parameters
        tp = TrainingParam()

        # this will be the list of what part of the observation I want to keep
        # more information on https://grid2op.readthedocs.io/en/latest/observation.html#main-observation-attributes
        li_attr_obs_X = ["day_of_week", "hour_of_day", "minute_of_hour", "prod_p", "prod_v", "load_p", "load_q",
                         "actual_dispatch", "target_dispatch", "topo_vect", "time_before_cooldown_line",
                         "time_before_cooldown_sub", "rho", "timestep_overflow", "line_status"]

        # neural network architecture
        observation_size = NNParam.get_obs_size(env, li_attr_obs_X)
        sizes = [800, 800, 800, 494, 494, 494]  # sizes of each hidden layers
        kwargs_archi = {'observation_size': observation_size,
                        'sizes': sizes,
                        'activs': ["relu" for _ in range(sizes)],  # all relu activation function
                        "list_attr_obs": li_attr_obs_X}

        # select some part of the action
        # more information at https://grid2op.readthedocs.io/en/latest/converter.html#grid2op.Converter.IdToAct.init_converter
        kwargs_converters = {"all_actions": None,
                             "set_line_status": False,
                             "change_bus_vect": True,
                             "set_topo_vect": False
                             }
        # define the name of the model
        nm_ = "AnneOnymous"
        try:
            train(env,
                  name=nm_,
                  iterations=10000,
                  save_path="/WHERE/I/SAVED/THE/MODEL",
                  load_path=None,
                  logs_dir="/WHERE/I/SAVED/THE/LOGS",
                  training_param=tp,
                  kwargs_converters=kwargs_converters,
                  kwargs_archi=kwargs_archi)
        finally:
            env.close()

    """

    # Limit gpu usage
    try:
        physical_devices = tf.config.list_physical_devices('GPU')
        if len(physical_devices) > 0:
            tf.config.experimental.set_memory_growth(physical_devices[0], True)
    except AttributeError:
         # issue of https://stackoverflow.com/questions/59266150/attributeerror-module-tensorflow-core-api-v2-config-has-no-attribute-list-p
        try:
            physical_devices = tf.config.experimental.list_physical_devices('GPU')
            if len(physical_devices) > 0:
                tf.config.experimental.set_memory_growth(physical_devices[0], True)
        except Exception:
            warnings.warn(_WARN_GPU_MEMORY)
    except Exception:
        warnings.warn(_WARN_GPU_MEMORY)

    if training_param is None:
        training_param = TrainingParam()

    # compute the proper size for the converter
    kwargs_archi["action_size"] = DuelQSimple.get_action_size(env.action_space, filter_action_fun, kwargs_converters)

    if load_path is not None:
        path_model, path_target_model = DuelQ_NN.get_path_model(load_path, name)
        if verbose:
            print("INFO: Reloading a model, the architecture parameters will be ignored")
        nn_archi = DuelQ_NNParam.from_json(os.path.join(path_model, "nn_architecture.json"))
    else:
        nn_archi = DuelQ_NNParam(**kwargs_archi)

    baseline = DuelQSimple(action_space=env.action_space,
                            nn_archi=nn_archi,
                            name=name,
                            istraining=True,
                            verbose=verbose,
                            **kwargs_converters
                            )

    if load_path is not None:
        print("INFO: Reloading a model, training parameters will be ignored")
        baseline.load(load_path)
        training_param = baseline._training_param

    baseline.train(env,
                   iterations,
                   save_path=save_path,
                   logdir=logs_dir,
                   training_param=training_param)
    # as in our example (and in our explanation) we recommend to save the mode regurlarly in the "train" function
    # it is not necessary to save it again here. But if you chose not to follow these advice, it is more than
    # recommended to save the "baseline" at the end of this function with:
    # baseline.save(path_save)


if __name__ == "__main__":
    # import grid2op
    import numpy as np
    from grid2op.Parameters import Parameters
    from grid2op import make
    from grid2op.Reward import L2RPNReward
    import re
    try:
        from lightsim2grid.LightSimBackend import LightSimBackend
        backend = LightSimBackend()
    except:
        from grid2op.Backend import PandaPowerBackend
        backend = PandaPowerBackend()

    args = cli_train().parse_args()

    # is it highly recommended to modify the reward depening on the algorithm.
    # for example here i will push my algorithm to learn that plyaing illegal or ambiguous action is bad
    class MyReward(L2RPNReward):
        def initialize(self, env):
            self.reward_min = 0.0
            self.reward_max = 1.0

        def __call__(self, action, env, has_error, is_done, is_illegal, is_ambiguous):
            if has_error or is_illegal or is_ambiguous:
                # previous action was bad
                res = self.reward_min
            elif is_done:
                # really strong reward if an episode is over without game over
                res = self.reward_max
            else:
                res = super().__call__(action, env, has_error, is_done, is_illegal, is_ambiguous)
                res /= env.n_line
                if not np.isfinite(res):
                    res = self.reward_min
            return res

    # Use custom params

    # Create grid2op game environement
    env_init = None
    from grid2op.Chronics import MultifolderWithCache
    game_param = Parameters()
    game_param.NB_TIMESTEP_COOLDOWN_SUB = 2
    game_param.NB_TIMESTEP_COOLDOWN_LINE = 2
    env = make(args.env_name,
               param=game_param,
               reward_class=MyReward,
               backend=backend,
               chronics_class=MultifolderWithCache
               )
    # env.chronics_handler.set_max_iter(7*288)
    env.chronics_handler.real_data.set_filter(lambda x: re.match(".*((0003)|(0072)|(0057))$", x) is not None)
    env.chronics_handler.real_data.reset_cache()
    # env.chronics_handler.real_data.
    env_init = env
    if args.nb_env > 1:
        from grid2op.Environment import MultiEnvironment
        env = MultiEnvironment(int(args.nb_env), env)
        # TODO hack i'll fix in 0.9.0
        env.action_space = env_init.action_space
        env.observation_space = env_init.observation_space
        env.fast_forward_chronics = lambda x: None
        env.chronics_handler = env_init.chronics_handler
        env.current_obs = env_init.current_obs
        env.set_ff()

    tp = TrainingParam()

    # NN training
    tp.lr = 1e-4
    tp.lr_decay_steps = 30000
    tp.minibatch_size = 256
    tp.update_freq = 128

    # limit the number of time steps played per scenarios
    tp.step_increase_nb_iter = 2
    tp.min_iter = 10
    tp.update_nb_iter(2)

    # oversampling hard scenarios
    tp.oversampling_rate = 3

    # experience replay
    tp.buffer_size = 1000000

    # e greedy
    tp.min_observation = 10000
    tp.initial_epsilon = 0.4
    tp.final_epsilon = 1./(2*7*288.)
    tp.step_for_final_epsilon = int(1e5)

    # don't start always at the same hour (if not None) otherwise random sampling, see docs
    tp.random_sample_datetime_start = None

    # saving, logging etc.
    tp.save_model_each = 10000
    tp.update_tensorboard_freq = 256

    li_attr_obs_X = ["day_of_week", "hour_of_day", "minute_of_hour", "prod_p", "prod_v", "load_p", "load_q",
                     "actual_dispatch", "target_dispatch", "topo_vect", "time_before_cooldown_line",
                     "time_before_cooldown_sub", "rho", "timestep_overflow", "line_status"]

    # nn architecture
    observation_size = DuelQ_NNParam.get_obs_size(env_init, li_attr_obs_X)

    kwargs_archi = {'observation_size': observation_size,
                    'sizes': [800, 800, 800, 494, 494, 494],
                    'activs': ["relu" for _ in range(6)],
                    "list_attr_obs": li_attr_obs_X}

    # which actions i keep
    kwargs_converters = {"all_actions": None,
                         "set_line_status": False,
                         "change_bus_vect": True,
                         "set_topo_vect": False
                         }
    nm_ = args.name if args.name is not None else DEFAULT_NAME
    try:
        train(env,
              name=nm_,
              iterations=args.num_train_steps,
              save_path=args.save_path,
              load_path=args.load_path,
              logs_dir=args.logs_dir,
              nb_env=args.nb_env,
              training_param=tp,
              kwargs_converters=kwargs_converters,
              kwargs_archi=kwargs_archi)
    finally:
        env.close()
        if args.nb_env > 1:
            env_init.close()



# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

import numpy as np
import os
import tensorflow as tf

# tf2.0 friendly
import warnings

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=FutureWarning)
    from tensorflow.keras.models import load_model, Sequential, Model
    import tensorflow.keras.optimizers as tfko
    from tensorflow.keras.layers import Activation, Dense
    from tensorflow.keras.layers import Input, Concatenate

from l2rpn_baselines.utils import BaseDeepQ, TrainingParam


# This class implements the "Sof Actor Critic" model.
# It is a custom implementation, courtesy to Clement Goubet
# The original paper is: https://arxiv.org/abs/1801.01290
class SAC_NN(BaseDeepQ):
    """
    Constructs the desired soft actor critic network.

    Compared to other baselines shown elsewhere (*eg* :class:`l2rpn_baselines.DeepQSimple` or
    :class:`l2rpn_baselines.DeepQSimple`) the implementation of the SAC is a bit more tricky.

    However, we demonstrate here that the use of :class:`l2rpn_baselines.utils.BaseDeepQ` with custom
    parameters class (in this calse :class:`SAC_NNParam` is flexible enough to meet our needs.

    """
    def __init__(self,
                 nn_params,
                 training_param=None,
                 verbose=False):
        if training_param is None:
            training_param = TrainingParam()
        BaseDeepQ.__init__(self,
                           nn_params,
                           training_param,
                           verbose=verbose)

        # TODO add as meta param the number of "Q" you want to use (here 2)
        # TODO add as meta param size and types of the networks
        self.average_reward = 0
        self.life_spent = 1
        self.qvalue_evolution = np.zeros((0,))
        self.Is_nan = False

        self.model_value_target = None
        self.model_value = None
        self.model_Q = None
        self.model_Q2 = None
        self.model_policy = None

        self.construct_q_network()
        self.previous_size = 0
        self.previous_eyes = None
        self.previous_arange = None
        self.previous_size_train = 0
        self.previous_eyes_train = None

        # optimizers and learning rate
        self.schedule_lr_policy = None
        self.optimizer_policy = None
        self.schedule_lr_Q = None
        self.optimizer_Q = None
        self.schedule_lr_Q2 = None
        self.optimizer_Q2 = None
        self.schedule_lr_value = None
        self.optimizer_value = None

    def _build_q_NN(self):
        input_states = Input(shape=(self._observation_size,))
        input_action = Input(shape=(self._action_size,))

        input_layer = Concatenate()([input_states, input_action])
        lay = input_layer
        for lay_num, (size, act) in enumerate(zip(self._nn_archi.sizes, self._nn_archi.activs)):
            lay = Dense(size, name="layer_{}".format(lay_num))(lay)  # put at self.action_size
            lay = Activation(act)(lay)

        advantage = Dense(1, activation='linear')(lay)

        model = Model(inputs=[input_states, input_action], outputs=[advantage])
        return model

    def _build_model_value(self):
        input_states = Input(shape=(self._observation_size,))

        lay = input_states
        for lay_num, (size, act) in enumerate(zip(self._nn_archi.sizes_value, self._nn_archi.activs_value)):
            lay = Dense(size)(lay)
            lay = Activation(act)(lay)

        advantage = Dense(self._action_size, activation='relu')(lay)
        state_value = Dense(1, activation='linear', name="state_value")(advantage)
        model = Model(inputs=[input_states], outputs=[state_value])
        return model

    def construct_q_network(self):
        """
        This constructs all the networks needed for the SAC agent.
        """
        self.model_Q = self._build_q_NN()
        self.schedule_lr_Q, self.optimizer_Q = self.make_optimiser()
        self.model_Q.compile(loss='mse', optimizer=self.optimizer_Q)

        self.model_Q2 = self._build_q_NN()
        self.schedule_lr_Q2, self.optimizer_Q2 = self.make_optimiser()
        self.model_Q2.compile(loss='mse', optimizer=self.optimizer_Q2)

        # state value function approximation
        self.model_value = self._build_model_value()
        self.schedule_lr_value, self.optimizer_value = self.make_optimiser()
        self._optimizer_model = self.optimizer_value
        self.model_value.compile(loss='mse', optimizer=self.optimizer_value)

        self.model_value_target = self._build_model_value()
        self.model_value_target.set_weights(self.model_value.get_weights())

        # policy function approximation
        self.model_policy = Sequential()
        # proba of choosing action a depending on policy pi
        input_states = Input(shape=(self._observation_size,))
        lay = input_states
        for lay_num, (size, act) in enumerate(zip(self._nn_archi.sizes_policy, self._nn_archi.activs_policy)):
            lay = Dense(size)(lay)
            lay = Activation(act)(lay)
        soft_proba = Dense(self._action_size, activation="softmax", kernel_initializer='uniform', name="soft_proba")(lay)
        self.model_policy = Model(inputs=[input_states], outputs=[soft_proba])
        self.schedule_lr_policy, self.optimizer_policy = self.make_optimiser()
        self.model_policy.compile(loss='categorical_crossentropy', optimizer=self.optimizer_policy)

    def _get_eye_pm(self, batch_size):
        if batch_size != self.previous_size:
            tmp = np.zeros((batch_size, self._action_size), dtype=np.float32)
            self.previous_eyes = tmp
            self.previous_arange = np.arange(batch_size)
            self.previous_size = batch_size
        return self.previous_eyes, self.previous_arange

    def predict_movement(self, data, epsilon, batch_size=None):
        """
        predict the next movements in a vectorized fashion
        """
        if batch_size is None:
            batch_size = data.shape[0]
        rand_val = np.random.random(data.shape[0])
        p_actions = self.model_policy.predict(data, batch_size=batch_size)
        opt_policy_orig = np.argmax(np.abs(p_actions), axis=-1)
        opt_policy = 1.0 * opt_policy_orig
        opt_policy[rand_val < epsilon] = np.random.randint(0, self._action_size, size=(np.sum(rand_val < epsilon)))
        opt_policy = opt_policy.astype(np.int)
        return opt_policy, p_actions[:, opt_policy]

    def _get_eye_train(self, batch_size):
        if batch_size != self.previous_size_train:
            self.previous_eyes_train = np.repeat(np.eye(self._action_size),
                                                 batch_size * np.ones(self._action_size, dtype=np.int),
                                                 axis=0)
            self.previous_eyes_train = tf.convert_to_tensor(self.previous_eyes_train, dtype=tf.float32)
            self.previous_size_train = batch_size
        return self.previous_eyes_train

    def train(self, s_batch, a_batch, r_batch, d_batch, s2_batch, tf_writer=None, batch_size=None):
        """Trains networks to fit given parameters"""
        if batch_size is None:
            batch_size = s_batch.shape[0]
        target = np.zeros((batch_size, 1))
        # training of the action state value networks
        last_action = np.zeros((batch_size, self._action_size))
        # Save the graph just the first time
        if tf_writer is not None:
            tf.summary.trace_on()
        fut_action = self.model_value_target.predict(s2_batch, batch_size=batch_size).reshape(-1)
        if tf_writer is not None:
            with tf_writer.as_default():
                tf.summary.trace_export("model_value_target-graph", 0)
            tf.summary.trace_off()

        target[:, 0] = r_batch + (1 - d_batch) * self._training_param.discount_factor * fut_action
        loss = self.model_Q.train_on_batch([s_batch, last_action], target)
        loss_2 = self.model_Q2.train_on_batch([s_batch, last_action], target)

        self.life_spent += 1
        temp = 1 / np.log(self.life_spent) / 2
        tiled_batch = np.tile(s_batch, (self._action_size, 1))
        tiled_batch_ts = tf.convert_to_tensor(tiled_batch)
        # tiled_batch: output something like: batch, batch, batch
        # TODO save that somewhere not to compute it each time, you can even save this in the
        # TODO tensorflow graph!
        tmp = self._get_eye_train(batch_size)

        action_v1_orig = self.model_Q.predict([tiled_batch_ts, tmp], batch_size=batch_size).reshape(batch_size, -1)
        action_v2_orig = self.model_Q2.predict([tiled_batch_ts, tmp], batch_size=batch_size).reshape(batch_size, -1)
        action_v1 = action_v1_orig - np.amax(action_v1_orig, axis=-1).reshape(batch_size, 1)
        new_proba = np.exp(action_v1 / temp) / np.sum(np.exp(action_v1 / temp), axis=-1).reshape(batch_size, 1)
        new_proba_ts = tf.convert_to_tensor(new_proba)
        loss_policy = self.model_policy.train_on_batch(s_batch, new_proba_ts)

        target_pi = self.model_policy.predict(s_batch, batch_size=batch_size)
        value_target = np.fmin(action_v1_orig[0, a_batch], action_v2_orig[0, a_batch]) - np.sum(
            target_pi * np.log(target_pi + 1e-6))
        value_target_ts = tf.convert_to_tensor(value_target.reshape(-1, 1))
        loss_value = self.model_value.train_on_batch(s_batch, value_target_ts)

        self.Is_nan = np.isnan(loss) + np.isnan(loss_2) + np.isnan(loss_policy) + np.isnan(loss_value)
        return np.all(np.isfinite(loss)) & np.all(np.isfinite(loss_2)) & np.all(np.isfinite(loss_policy)) & \
               np.all(np.isfinite(loss_value))

    @staticmethod
    def _get_path_model(path, name=None):
        if name is None:
            path_model = path
        else:
            path_model = os.path.join(path, name)
        path_target_model = "{}_target".format(path_model)
        path_modelQ = "{}_Q".format(path_model)
        path_modelQ2 = "{}_Q2".format(path_model)
        path_policy = "{}_policy".format(path_model)
        return path_model, path_target_model, path_modelQ, path_modelQ2, path_policy

    def save_network(self, path, name=None, ext="h5"):
        """
        Saves all the models with unique names
        """
        path_model, path_target_model, path_modelQ, path_modelQ2, path_policy = self._get_path_model(path, name)
        self.model_value.save('{}.{}'.format(path_model, ext))
        self.model_value_target.save('{}.{}'.format(path_target_model, ext))
        self.model_Q.save('{}.{}'.format(path_modelQ, ext))
        self.model_Q2.save('{}.{}'.format(path_modelQ2, ext))
        self.model_policy.save('{}.{}'.format(path_policy, ext))

    def load_network(self, path, name=None, ext="h5"):
        """
        We load all the models using the keras "load_model" function.
        """
        path_model, path_target_model, path_modelQ, path_modelQ2, path_policy = self._get_path_model(path, name)
        self.model_value = load_model('{}.{}'.format(path_model, ext))
        self.model_value_target = load_model('{}.{}'.format(path_target_model, ext))
        self.model_Q = load_model('{}.{}'.format(path_modelQ, ext))
        self.model_Q2 = load_model('{}.{}'.format(path_modelQ2, ext))
        self.model_policy = load_model('{}.{}'.format(path_policy, ext))
        if self.verbose:
            print("Succesfully loaded network.")

    def target_train(self):
        """
        This update the target model.
        """
        model_weights = self.model_value.get_weights()
        target_model_weights = self.model_value_target.get_weights()
        for i in range(len(model_weights)):
            target_model_weights[i] = self._training_param.tau * model_weights[i] + (1 - self._training_param.tau) * \
                                      target_model_weights[i]
        self.model_value_target.set_weights(model_weights)



__all__ = [
    "SliceRDQN",
    "SliceRDQN_Config",
    "evaluate",
    "train"
]

from l2rpn_baselines.SliceRDQN.SliceRDQN import SliceRDQN
from l2rpn_baselines.SliceRDQN.SliceRDQN_Config import SliceRDQN_Config
from l2rpn_baselines.SliceRDQN.evaluate import evaluate
from l2rpn_baselines.SliceRDQN.train import train




try:
    from grid2op import make
    from grid2op.Reward import L2RPNReward
    from grid2op.Reward import CombinedReward, CloseToOverflowReward, GameplayReward
    from grid2op.Observation import CompleteObservation
    import numpy as np
    import os

    from l2rpn_baselines.AsynchronousActorCritic.Runner import run
    from l2rpn_baselines.AsynchronousActorCritic.AsynchronousActorCritic import A3CAgent
    from l2rpn_baselines.AsynchronousActorCritic.Action_reduced_list import main_function
except ImportError as exc_:
    raise ImportError("AsynchronousActorCritic baseline impossible to load the required dependencies for testing the model. The error was: \n {}".format(exc_))

# import Runner
# import Action_reduced_list
# from ActorCritic_Agent import *

def useful_state(obs,env):
    selected_obs = np.hstack((obs.topo_vect,obs.line_status)) #  should it not be nan for lines/gens that are disconnected ??
    selected_obs = np.hstack((selected_obs,obs.load_p/100))#
    selected_obs = np.hstack((selected_obs,obs.load_q/100))
    selected_obs = np.hstack((selected_obs,obs.prod_p/100))
    selected_obs = np.hstack((selected_obs,obs.prod_v/env.backend.prod_pu_to_kv))
    selected_obs = np.hstack((selected_obs,obs.rho))
    # selected_obs = np.hstack((selected_obs,obs.day))
    selected_obs = np.hstack((selected_obs,obs.hour_of_day/24))
    selected_obs = np.hstack((selected_obs,obs.minute_of_hour/60))
    # selected_obs = np.hstack((selected_obs,obs.day_of_week/7))
    return selected_obs

def evaluate(env,
                 load_path="",
                 logs_path=None,
                 nb_episode=1,
                 nb_process=None,
                 max_steps=-1,
                 verbose=False,
                 save_gif=False,
                 **kwargs):
        # env:grid2op.Environment.Environment The environment on which the baseline will be evaluated.
        # load_path: str The path where the model is stored. This is used by the agent when calling agent.load
        # logs_path: str The path where the agents results will be stored.
        # nb_episode: int Number of episodes to run for the assessment of the performance.
        # nb_process: int Number of process to be used for the assessment of the performance. NOTE: Not supported for this A3C code
        # max_steps: int Maximum number of timestep each episode can last. It should be a positive integer or -1. -1 means that the entire episode is run (until the chronics is out of data or until a game over).
        # verbose: bool verbosity of the output
        # save_gif: bool Whether or not to save a gif into each episode folder corresponding to the representation of the said episode.
        # agent_name: Name of the agent that is being tested. This will reflect inside the "log_path".
        # discrete_action_space: This is a list of lists which contain actions on power grid elements. A specific action can be accessed using the index of the list.
        # kwargs: Other key words arguments that you are free to use.
        # Create agent and load the trained model.

    runner_params = env.get_params_for_runner()
    if verbose:
        runner_params["verbose"] = True

    action_space_lists_data = kwargs["discrete_action_space"]
    NeuralNets_dimensions = kwargs["trained_NeuralNets_dimensions"]

    # for A3CAgent_for_runner file
    # agent = A3CAgent(env.action_space, state_size,action_size,env,action_space_lists=action_space_lists_data)

    # for ActorCritic_Agent.py file
    agent = A3CAgent(state_size, action_size,env.name,env.action_space,env.backend.prod_pu_to_kv,action_space_lists_data,
                            None,None,None,NeuralNets_dimensions,None,train_flag=False,save_path=None)

    # Print model summary
    agent.actor.summary()
    agent.critic.summary()

    # Name of agent folder where the performance data is stored.
    agent_name = kwargs["agent_name"]

    # Load the agent with trained neural network weights.
    try:
        agent.load_model(nn_weight_name=kwargs["nn_file_name"], load_path=load_path)
        print("Loaded saved NN model parameters \n")
    except:
        print("Issue with loading the NN weights/No existing model is found or saved model sizes do not match. Exiting...\n")
        exit()

    # Evaluate and print results and also save the gif if opted.
    run(runner_params,nb_episode,logs_path,max_steps,agent,save_gif,agent_name)
if __name__ == "__main__":

    NB_EPISODE = 1
    agent_log_path = "agents_log" # name of the folder to save the performance of the agent
    agent_nn_weights_name = "grid2op_14_a3c"
    name_of_RL_agent = "agent_1"
    max_steps_in_episode = 3000
    Hyperparameters = {}
    Hyperparameters["size_of_hidden_layer_1"] = 200
    Hyperparameters["size_of_hidden_layer_2"] = 100
    load_path = os.path.join(os.getcwd(),"nn_weight_folder")

    # Create environment
    env = make("l2rpn_case14_sandbox",reward_class=L2RPNReward)

    # env = make("l2rpn_case14_sandbox",reward_class=CombinedReward)
    # # Register custom reward for training
    # cr = env.reward_helper.template_reward
    # cr.addReward("overflow", CloseToOverflowReward(), 50.0)
    # cr.addReward("game", GameplayReward(), 100.0)
    # cr.initialize(env)

    # Get the size of state space and action space.
    do_nothing_act = env.helper_action_player({})
    obs, reward, done, info = env.step(do_nothing_act)
    state_trimmed = useful_state(obs,env)
    state_trimmed = state_trimmed.reshape([1,state_trimmed.size])
    state_size = state_trimmed.size

    # NOTE: Delete the .npy files if you are trying to solve for different bus system than previous solved power system
    # This code is implemented for a 14 bus system.
    # Reduce the action space and assign an index value to each unique possible action from the complete action space.
    gen_action_list, load_action_list, line_or_action_list, line_ex_action_list = main_function("14")
    action_size = load_action_list.__len__()
    del env

    action_space_lists = [gen_action_list, load_action_list, line_or_action_list, line_ex_action_list]

    env = make("l2rpn_case14_sandbox",reward_class=L2RPNReward)

    # env = make("l2rpn_case14_sandbox",reward_class=CombinedReward)
    # # Register custom reward for training
    # cr = env.reward_helper.template_reward
    # cr.addReward("overflow", CloseToOverflowReward(), 50.0)
    # cr.addReward("game", GameplayReward(), 100.0)
    # cr.initialize(env)

    # Debug prints
    print("Debug prints --->:")
    print("Chronics location that being used:", env.chronics_handler.path)
    print("Grid location being used:", env.init_grid_path)
    print("Reward class that is being used:", env.rewardClass)
    print("Action type class being used:", env.actionClass)
    print("Observation type class being used:", env.observationClass)
    print("Backend CSV file key names:", env.names_chronics_to_backend)
    print("Legal action class being used:", env.legalActClass)
    print("Voltage controller class being used:", env.voltagecontrolerClass)

    evaluate(env=env,
             load_path=load_path,
             logs_path=agent_log_path,
             nb_episode=NB_EPISODE,
             max_steps=max_steps_in_episode,
             verbose=True,
             save_gif=False,
             agent_name=name_of_RL_agent,
             discrete_action_space=action_space_lists,
             trained_NeuralNets_dimensions=Hyperparameters,
             nn_file_name=agent_nn_weights_name)



# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

import numpy as np
from l2rpn_baselines.utils import DeepQAgent
from l2rpn_baselines.DuelQLeapNet.DuelQLeapNet_NN import DuelQLeapNet_NN

DEFAULT_NAME = "DuelQLeapNet"


class DuelQLeapNet(DeepQAgent):
    def __init__(self,
                 action_space,
                 nn_archi,
                 name="DeepQAgent",
                 store_action=True,
                 istraining=False,
                 nb_env=1,
                 **kwargs_converters):
        DeepQAgent.__init__(self,
                            action_space,
                            nn_archi,
                            name=name,
                            store_action=store_action,
                            istraining=istraining,
                            nb_env=nb_env,
                            **kwargs_converters)
        self.tau_dim_start = None
        self.tau_dim_end = None
        self.add_tau = -1  # remove one to tau to have a vector of 0 and 1 instead of 1 and 2
        self._tmp_obs = None



# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

import os
import json
import copy
import numpy as np
import tensorflow as tf

from grid2op.Parameters import Parameters
from grid2op.Agent import AgentWithConverter
from grid2op.Converter import IdToAct

from l2rpn_baselines.DoubleDuelingRDQN.DoubleDuelingRDQNConfig import DoubleDuelingRDQNConfig as cfg
from l2rpn_baselines.DoubleDuelingRDQN.ExperienceBuffer import ExperienceBuffer
from l2rpn_baselines.DoubleDuelingRDQN.DoubleDuelingRDQN_NN import DoubleDuelingRDQN_NN

class DoubleDuelingRDQN(AgentWithConverter):
    def __init__(self,
                 observation_space,
                 action_space,
                 name=__name__,
                 is_training=False):
        # Call parent constructor
        AgentWithConverter.__init__(self, action_space,
                                    action_space_converter=IdToAct)

        # Store constructor params
        self.observation_space = observation_space
        self.name = name
        self.trace_length = cfg.TRACE_LENGTH
        self.batch_size = cfg.BATCH_SIZE
        self.is_training = is_training
        self.lr = cfg.LR
        
        # Declare required vars
        self.Qmain = None
        self.obs = None
        self.state = []
        self.mem_state = None
        self.carry_state = None

        # Declare training vars
        self.exp_buffer = None
        self.done = False
        self.epoch_rewards = None
        self.epoch_alive = None
        self.Qtarget = None

        # Compute dimensions from intial state
        self.observation_size = self.observation_space.size_obs()
        self.action_size = self.action_space.size()

        # Load network graph
        self.Qmain = DoubleDuelingRDQN_NN(self.action_size,
                                          self.observation_size,
                                          learning_rate = self.lr)
        # Setup training vars if needed
        if self.is_training:
            self._init_training()


    def _init_training(self):
        self.exp_buffer = ExperienceBuffer(cfg.REPLAY_BUFFER_SIZE,
                                           self.batch_size,
                                           self.trace_length)
        self.done = True
        self.epoch_rewards = []
        self.epoch_alive = []
        self.Qtarget = DoubleDuelingRDQN_NN(self.action_size,
                                            self.observation_size,
                                            learning_rate = self.lr)

    def _reset_state(self, current_obs):
        # Initial state
        self.obs = current_obs
        self.state = self.convert_obs(self.obs)
        self.done = False
        self.mem_state = np.zeros(self.Qmain.h_size)
        self.carry_state = np.zeros(self.Qmain.h_size)

    def _register_experience(self, episode_exp, episode):
        missing_obs = self.trace_length - len(episode_exp)

        if missing_obs > 0: # We are missing exp to make a trace
            exp = episode_exp[0] # Use inital state to fill out
            for missing in range(missing_obs):
                # Use do_nothing action at index 0
                self.exp_buffer.add(exp[0], 0, exp[2], exp[3], exp[4], episode)

        # Register the actual experience
        for exp in episode_exp:
            self.exp_buffer.add(exp[0], exp[1], exp[2], exp[3], exp[4], episode)

    def _save_hyperparameters(self, logpath, env, steps):
        r_instance = env.reward_helper.template_reward
        hp = {
            "lr": cfg.LR,
            "batch_size": cfg.BATCH_SIZE,
            "trace_len": cfg.TRACE_LENGTH,
            "e_start": cfg.INITIAL_EPSILON,
            "e_end": cfg.FINAL_EPSILON,
            "e_decay": cfg.DECAY_EPSILON,
            "discount": cfg.DISCOUNT_FACTOR,
            "buffer_size": cfg.REPLAY_BUFFER_SIZE,
            "update_freq": cfg.UPDATE_FREQ,
            "update_hard": cfg.UPDATE_TARGET_HARD_FREQ,
            "update_soft": cfg.UPDATE_TARGET_SOFT_TAU,
            "reward": dict(r_instance)
        }
        hp_filename = "{}-hypers.json".format(self.name)
        hp_path = os.path.join(logpath, hp_filename)
        with open(hp_path, 'w') as fp:
            json.dump(hp, fp=fp, indent=2)

    ## Agent Interface
    def convert_obs(self, observation):
        # Made a custom version to normalize per attribute
        #return observation.to_vect()
        li_vect=  []
        for el in observation.attr_list_vect:
            v = observation._get_array_from_attr_name(el).astype(np.float)
            v_fix = np.nan_to_num(v)
            v_norm = np.linalg.norm(v_fix)
            if v_norm > 1e6:
                v_res = (v_fix / v_norm) * 10.0
            else:
                v_res = v_fix
            li_vect.append(v_res)
        return np.concatenate(li_vect)

    def convert_act(self, action):
        return super().convert_act(action)

    def reset(self, observation):
        self._reset_state(observation)

    def my_act(self, state, reward, done=False):
        data_input = np.array(state)
        data_input.reshape(1, 1, self.observation_size)
        a, _, m, c = self.Qmain.predict_move(data_input,
                                             self.mem_state,
                                             self.carry_state)
        self.mem_state = m
        self.carry_state = c

        return a
    
    def load(self, path):
        self.Qmain.load_network(path)
        if self.is_training:
            self.Qmain.update_target_hard(self.Qtarget.model)

    def save(self, path):
        self.Qmain.save_network(path)

    ## Training Procedure
    def train(self, env,
              iterations,
              save_path,
              num_pre_training_steps = 0,
              logdir = "logs"):

        # Loop vars
        num_training_steps = iterations
        num_steps = num_pre_training_steps + num_training_steps
        step = 0
        epsilon = cfg.INITIAL_EPSILON
        alive_steps = 0
        total_reward = 0
        episode = 0
        episode_exp = []

        # Create file system related vars
        logpath = os.path.join(logdir, self.name)
        os.makedirs(save_path, exist_ok=True)
        modelpath = os.path.join(save_path, self.name + ".tf")
        self.tf_writer = tf.summary.create_file_writer(logpath, name=self.name)
        self._save_hyperparameters(save_path, env, num_steps)
        
        # Training loop
        self._reset_state(env.current_obs)
        while step < num_steps:
            # New episode
            if self.done:
                new_obs = env.reset() # This shouldn't raise
                self._reset_state(new_obs)
                # Push current episode experience to experience buffer
                self._register_experience(episode_exp, episode)
                # Reset current episode experience
                episode += 1
                episode_exp = []

            if cfg.VERBOSE and step % 1000 == 0:
                print("Step [{}] -- Dropout [{}]".format(step, epsilon))

            # Choose an action
            if step <= num_pre_training_steps:
                a, m, c = self.Qmain.random_move(self.state,
                                                 self.mem_state,
                                                 self.carry_state)
            elif len(episode_exp) < self.trace_length:
                a, m, c = self.Qmain.random_move(self.state,
                                                 self.mem_state,
                                                 self.carry_state)
                a = 0 # Do Nothing
            else:
                a, _, m, c = self.Qmain.bayesian_move(self.state,
                                                      self.mem_state,
                                                      self.carry_state,
                                                      epsilon)

            # Update LSTM state
            self.mem_state = m
            self.carry_state = c

            # Convert it to a valid action
            act = self.convert_act(a)
            # Execute action
            new_obs, reward, self.done, info = env.step(act)
            new_state = self.convert_obs(new_obs)
            
            # Save to current episode experience
            episode_exp.append((self.state, a, reward, self.done, new_state))

            # Train when pre-training is over
            if step >= num_pre_training_steps:
                training_step = step - num_pre_training_steps
                # Slowly decay dropout rate
                if epsilon > cfg.FINAL_EPSILON:
                    epsilon -= cfg.STEP_EPSILON
                if epsilon < cfg.FINAL_EPSILON:
                    epsilon = cfg.FINAL_EPSILON

                # Perform training at given frequency
                if step % cfg.UPDATE_FREQ == 0 and \
                   self.exp_buffer.can_sample():
                    # Sample from experience buffer
                    batch = self.exp_buffer.sample()
                    # Perform training
                    self._batch_train(batch, step, training_step)
                    # Update target network towards primary network
                    if cfg.UPDATE_TARGET_SOFT_TAU > 0:
                        tau = cfg.UPDATE_TARGET_SOFT_TAU
                        self.Qmain.update_target_soft(self.Qtarget.model, tau)

                # Every UPDATE_TARGET_HARD_FREQ trainings,
                # update target completely
                if cfg.UPDATE_TARGET_HARD_FREQ > 0 and \
                   step % (cfg.UPDATE_FREQ * cfg.UPDATE_TARGET_HARD_FREQ) == 0:
                    self.Qmain.update_target_hard(self.Qtarget.model)

            total_reward += reward
            if self.done:
                self.epoch_rewards.append(total_reward)
                self.epoch_alive.append(alive_steps)
                if cfg.VERBOSE:
                    print("Survived [{}] steps".format(alive_steps))
                    print("Total reward [{}]".format(total_reward))
                alive_steps = 0
                total_reward = 0
            else:
                alive_steps += 1
            
            # Save the network every 1000 iterations
            if step > 0 and step % 1000 == 0:
                self.save(modelpath)

            # Iterate to next loop
            step += 1
            self.obs = new_obs
            self.state = new_state

        # Save model after all steps
        self.save(modelpath)

    def _batch_train(self, batch, step, training_step):
        """Trains network to fit given parameters"""
        Q = np.zeros((self.batch_size, self.action_size))
        batch_mem = np.zeros((self.batch_size, self.Qmain.h_size))
        batch_carry = np.zeros((self.batch_size, self.Qmain.h_size))

        input_size = self.observation_size
        m_data = np.vstack(batch[:, 0])
        m_data = m_data.reshape(self.batch_size, self.trace_length, input_size)
        t_data = np.vstack(batch[:, 4])
        t_data = t_data.reshape(self.batch_size, self.trace_length, input_size)
        q_input = [
            copy.deepcopy(batch_mem),
            copy.deepcopy(batch_carry),
            copy.deepcopy(m_data)
        ]
        q1_input = [
            copy.deepcopy(batch_mem),
            copy.deepcopy(batch_carry),
            copy.deepcopy(t_data)
        ]
        q2_input = [
            copy.deepcopy(batch_mem),
            copy.deepcopy(batch_carry),
            copy.deepcopy(t_data)
        ]

        # Batch predict
        self.Qmain.trace_length.assign(self.trace_length)
        self.Qmain.dropout_rate.assign(0.0)
        self.Qtarget.trace_length.assign(self.trace_length)
        self.Qtarget.dropout_rate.assign(0.0)

        # Save the graph just the first time
        if training_step == 0:
            tf.summary.trace_on()

        # T Batch predict
        Q, _, _ = self.Qmain.model.predict(q_input,
                                           batch_size = self.batch_size)

        ## Log graph once and disable graph logging
        if training_step == 0:
            with self.tf_writer.as_default():
                tf.summary.trace_export(self.name + "-graph", step)

        # T+1 batch predict
        Q1, _, _ = self.Qmain.model.predict(q1_input,
                                            batch_size=self.batch_size)
        Q2, _, _ = self.Qtarget.model.predict(q2_input,
                                              batch_size=self.batch_size)

        # Compute batch Double Q update to Qtarget
        for i in range(self.batch_size):
            idx = i * (self.trace_length - 1)
            doubleQ = Q2[i, np.argmax(Q1[i])]
            a = batch[idx][1]
            r = batch[idx][2]
            d = batch[idx][3]
            Q[i, a] = r
            if d == False:
                Q[i, a] += cfg.DISCOUNT_FACTOR * doubleQ

        # Batch train
        batch_x = [batch_mem, batch_carry, m_data]
        batch_y = [Q, batch_mem, batch_carry]
        loss = self.Qmain.model.train_on_batch(batch_x, batch_y)
        loss = loss[0]

        # Log some useful metrics
        if step % (cfg.UPDATE_FREQ * 2) == 0:
            if cfg.VERBOSE:
                print("loss =", loss)

            with self.tf_writer.as_default():
                mean_reward = np.mean(self.epoch_rewards)
                mean_alive = np.mean(self.epoch_alive)
                if len(self.epoch_rewards) >= 100:
                    mean_reward_100 = np.mean(self.epoch_rewards[-100:])
                    mean_alive_100 = np.mean(self.epoch_alive[-100:])
                else:
                    mean_reward_100 = mean_reward
                    mean_alive_100 = mean_alive
                tf.summary.scalar("mean_reward", mean_reward, step)
                tf.summary.scalar("mean_alive", mean_alive, step)
                tf.summary.scalar("mean_reward_100", mean_reward_100, step)
                tf.summary.scalar("mean_alive_100", mean_alive_100, step)
                tf.summary.scalar("loss", loss, step)



import numpy as np    

def lines_q_len(action_space):
    all_actions = [action_space({})]
    if "_set_line_status" in action_space._template_act.attr_list_vect:
        all_actions += action_space.get_all_unitary_line_set(action_space)

    if "_switch_line_status" in action_space._template_act.attr_list_vect:
        all_actions += action_space.get_all_unitary_line_change(action_space)

    return len(all_actions)

def topo_q_len(action_space):
    all_actions = [action_space({})]
    if "_set_topo_vect" in action_space._template_act.attr_list_vect:
        all_actions += action_space.get_all_unitary_topologies_set(action_space)

    if "_change_bus_vect" in action_space._template_act.attr_list_vect:
        all_actions += action_space.get_all_unitary_topologies_change(action_space)

    return len(all_actions)

def disp_q_len(action_space):
    all_actions = [action_space({})]
    if "_redispatch" in action_space._template_act.attr_list_vect:
        all_actions += action_space.get_all_unitary_redispatch(action_space)

    return len(all_actions)

def shape_obs(observation_space):
    dims = np.array([
        observation_space.n_line,
        observation_space.n_gen,
        observation_space.n_load,
        observation_space.n_sub,
        5
    ])
    return (25, np.max(dims))

def to_pad_vect(inputv, pad_w, pad_v = 0.0, scale_v = 1.0):
    v = np.asarray(inputv)
    v = v / scale_v
    vsafe = np.nan_to_num(v, nan=pad_v, posinf=pad_v, neginf=pad_v)
    padder = (0, pad_w - len(vsafe))
    vpad = np.pad(vsafe, padder, constant_values=pad_v)
    return vpad.astype(np.float32)

def convert_obs_pad(obs, bias=0.0):
    # Store some shortcuts
    topo = obs.topo_vect
    g_pos = obs.gen_pos_topo_vect
    l_pos = obs.load_pos_topo_vect
    lor_pos = obs.line_or_pos_topo_vect
    lex_pos = obs.line_ex_pos_topo_vect
    
    # Find longest dim
    dims = np.array([obs.n_line, obs.n_gen, obs.n_load, obs.n_sub, 5])
    pad_w = np.max(dims)

    # Get time data
    time_li = [obs.month / 12.0, obs.day / 31.0, obs.day_of_week / 7.0,
               obs.hour_of_day / 24.0, obs.minute_of_hour / 60.0]
    time_v = to_pad_vect(time_li, pad_w)
    time_line_cd = to_pad_vect(obs.time_before_cooldown_line, pad_w, pad_v=-1.0, scale_v=10.0)
    time_line_nm = to_pad_vect(obs.time_next_maintenance, pad_w, scale_v=10.0)
    time_sub_cd = to_pad_vect(obs.time_before_cooldown_sub, pad_w, pad_v=-1.0, scale_v=10.0)
    
    # Get generators info
    g_p = to_pad_vect(obs.prod_p, pad_w, scale_v=1000.0)
    g_q = to_pad_vect(obs.prod_q, pad_w, scale_v=1000.0)
    g_v = to_pad_vect(obs.prod_v, pad_w, scale_v=1000.0)
    g_tr = to_pad_vect(obs.target_dispatch, pad_w, scale_v=150.0)
    g_ar = to_pad_vect(obs.actual_dispatch, pad_w, scale_v=150.0)
    g_cost = to_pad_vect(obs.gen_cost_per_MW, pad_w, pad_v=0.0, scale_v=1.0)
    g_buses = np.zeros(obs.n_gen)
    for gen_id in range(obs.n_gen):
        g_buses[gen_id] = topo[g_pos[gen_id]]
        if g_buses[gen_id] <= 0.0:
            g_buses[gen_id] = 0.0
    g_bus = to_pad_vect(g_buses, pad_w, pad_v=-1.0, scale_v=3.0)

    # Get loads info
    l_p = to_pad_vect(obs.load_p, pad_w, scale_v=1000.0)
    l_q = to_pad_vect(obs.load_q, pad_w, scale_v=1000.0)
    l_v = to_pad_vect(obs.load_v, pad_w, scale_v=1000.0)
    l_buses = np.zeros(obs.n_load)
    for load_id in range(obs.n_load):
        l_buses[load_id] = topo[l_pos[load_id]]
        if l_buses[load_id] <= 0.0:
            l_buses[load_id] = 0.0
    l_bus = to_pad_vect(l_buses, pad_w, pad_v=-1.0, scale_v=3.0)

    # Get lines origin info
    or_p = to_pad_vect(obs.p_or, pad_w, scale_v=1000.0)
    or_q = to_pad_vect(obs.q_or, pad_w, scale_v=1000.0)
    or_v = to_pad_vect(obs.v_or, pad_w, scale_v=1000.0)
    or_buses = np.zeros(obs.n_line)
    for line_id in range(obs.n_line):
        or_buses[line_id] = topo[lor_pos[line_id]]
        if or_buses[line_id] <= 0.0:
            or_buses[line_id] = 0.0
    or_bus = to_pad_vect(or_buses, pad_w, pad_v=-1.0, scale_v=3.0)
    or_rho = to_pad_vect(obs.rho, pad_w, pad_v=-1.0)
    
    # Get extremities origin info
    ex_p = to_pad_vect(obs.p_ex, pad_w, scale_v=1000.0)
    ex_q = to_pad_vect(obs.q_ex, pad_w, scale_v=1000.0)
    ex_v = to_pad_vect(obs.v_ex, pad_w, scale_v=1000.0)
    ex_buses = np.zeros(obs.n_line)
    for line_id in range(obs.n_line):
        ex_buses[line_id] = topo[lex_pos[line_id]]
        if ex_buses[line_id] <= 0.0:
            ex_buses[line_id] = 0.0
    ex_bus = to_pad_vect(ex_buses, pad_w, pad_v=-1.0, scale_v=3.0)
    ex_rho = to_pad_vect(obs.rho, pad_w, pad_v=-1.0)

    res = np.stack([
        # [0;3] Time
        time_v, time_line_cd, time_sub_cd, time_line_nm,
        # [4;10] Gens
        g_p, g_q, g_v, g_ar, g_tr, g_bus, g_cost,
        # [11;14] Loads
        l_p, l_q, l_v, l_bus,
        # [15;19] Origins
        or_p, or_q, or_v, or_bus, or_rho,
        # [20;24] Extremities
        ex_p, ex_q, ex_v, ex_bus, ex_rho
    ])
    return res + bias



import os
from grid2op import make

# name of the folder to save the performance of the agent.
agent_log_path = "agents_log"
# Path where the grid2viz is downlaoded and installed using "git clone" and "pip install -U ."
grid2viz_location = f"C:Users\kisha\Grid2Viz\grid2viz"

# In my case, the agent .npz data is located in the same location as this .py file.
AGENTS_PATH = os.path.join(os.getcwd(),agent_log_path)
print("Current working directory:", os.getcwd())
print("Change the working directory to the grid2viz location:", grid2viz_location)
os.chdir(grid2viz_location)
print("Current working directory:", os.getcwd())

exit_code = os.system(f"python main.py --agents_path {AGENTS_PATH}")



# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

import os
from grid2op.Episode import EpisodeReplay


def save_log_gif(path_log, res, gif_name=None):
    """
    Output a gif named (by default "episode.gif") that is the replay of the episode in a gif format,
    for each episode in the input.

    Parameters
    ----------
    path_log: ``str``
        Path where the log of the agents are saved.

    res: ``list``
        List resulting from the call to `runner.run`

    gif_name: ``str``
        Name of the gif that will be used.

    """
    init_gif_name = gif_name
    ep_replay = EpisodeReplay(path_log)
    for _, chron_name, cum_reward, nb_time_step, max_ts in res:
        if gif_name is None:
            gif_name = chron_name
        gif_path = os.path.join(path_log, chron_name, gif_name)
        print("Creating {}.gif".format(gif_name))
        ep_replay.replay_episode(episode_id=chron_name, gif_name=gif_name, display=False)
        print("Wrote {}.gif".format(gif_path))
        gif_name = init_gif_name


if __name__ == "__main__":
    pass


# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

from grid2op.Agent import BaseAgent


class DoNothing(BaseAgent):
    """
    Do nothing agent of grid2op, as a lowerbond baseline for l2rpn competition.
    """
    def __init__(self,
                 action_space,
                 observation_space,
                 name,
                 **kwargs):
        super().__init__(action_space)
        self.name = name

    def act(self, observation, reward, done):
        # Just return an empty action aka. "do nothing"
        return self.action_space({})

    def reset(self, observation):
        # No internal states to reset
        pass

    def load(self, path):
        # Nothing to load
        pass

    def save(self, path):
        # Nothing to save
        pass




# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.
import os
import sys
import zipfile
import tempfile
import datetime
import warnings

import grid2op


def zip_for_codalab(path_agent, dest="."):
    """DEPRECATED DO NOT USE"""
    warnings.warn("This is a deprecated version of the zip file. Please avoid using it and use the one given in the "
                  "starting kit of the competition.")
    folder = os.path.abspath(path_agent)
    if not os.path.exists(folder):
        raise RuntimeError("The folder \"{}\" is empty and cannot be send to codalab to serve as a submitted agent.")

    root, directory_ = os.path.split(folder)
    sys.path.append(root)
    try:
        exec(f"from {directory_} import make_agent")
    except:
        raise RuntimeError(f"Impossible to import the \"make_agent\" function that is used by codalab to... create "
                           "your agent. Your submission is not valid. Please make sure the {path_agent} module "
                           "exposes the \"make_agent\" function.")

    try:
        exec(f"from {directory_} import reward")
    except:
        print("INFO: No custom reward for the assessment of your agent will be used. If you want to use a custom "
              "reward when your agent is evaluated, make sure to export  \"reward\", which should be a class "
              "inheriting from grid2op.BaseReward in your module (done in __init__.py).")

    try:
        exec(f"from {directory_} import other_rewards")
    except:
        print("INFO: No custom other_rewards for the assessment of your agent will be used. If you want to get "
              "information about other rewards when your agent is evaluated, make sure to export  \"other_rewards\" "
              "dictionnary in your module (you can do it in the __init__.py file)")

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        env_test = grid2op.make_new("rte_case5_example", test=True)
    try:
        toto = make_agent(env_test, folder)
    except TypeError:
        print("WARNING: a call to \"make_agent(environment, path_agent)\" raise a TypeError. There are great chances "
              "that your agent will not be valid on codalab. This is a warning but not an error, because the "
              "environment and the path_agent are 'fake' data for the test. If i were you i would double check "
              "though.")
    except:
        pass

    print("Your submission appear to be valid. For more test, we encourage you to run the appropriate notebook to "
          "do these unit testing.")
    the_date = datetime.datetime.now()
    zipped_submission = f'submission_{the_date:%y-%m-%d-%H-%M}.zip'
    fp = tempfile.TemporaryFile(mode="w", encoding="utf-8", dir=dest)
    fp.write('This is internal to codalab, do not modify')
    fp.seek(0)
    dest = os.path.abspath(dest)
    zip_file_name = os.path.join(dest, zipped_submission)
    with zipfile.ZipFile(zip_file_name, 'w', zipfile.ZIP_DEFLATED) as zipf:
        zipf.write(fp.name, arcname="metadata")
        # Do stuff here
        for root, dirs, files in os.walk(folder, topdown=True):
            dirs[:] = [d for d in dirs if d not in {"__pycache__"}]
            for file_ in files:
                if os.path.splitext(file_)[1] == ".pyc":
                    continue
                arc_path = os.path.relpath(os.path.join(root, file_), os.path.join(folder))
                arc_path = os.path.join("submission", arc_path)
                zipf.write(os.path.join(root, file_),
                           arcname=arc_path)
    print(f"The zip file \"{zip_file_name}\" has been created with your submission in it.")
    return zip_file_name


if __name__ == "__main__":
    pass


# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

import argparse


def cli_train():
    """some default command line arguments (cli) for training the baselines. Can be reused in some baselines here."""
    parser = argparse.ArgumentParser(description="Train baseline DDQN")
    parser.add_argument("--num_train_steps", required=False,
                        default=1024, type=int,
                        help="Number of training iterations")
    parser.add_argument("--save_path", required=False,
                        help="Path where the model should be saved.")
    parser.add_argument("--name", required=False,
                        help="Name given to your model.")
    parser.add_argument("--nb_env", required=False, default=1, type=int,
                        help="Number of process to use when training your Agent. If > 1 then MultiEnv will be used. "
                             "NB: not all models are compatible. NB it does not work on windows at the moment. "
                             "NB: experimental at the moment.")
    parser.add_argument("--load_path", required=False,
                        help="Path from which to reload your model from (by default ``None`` to NOT reload anything)")
    parser.add_argument("--env_name", required=False, default="l2rpn_case14_sandbox",
                        help="Name of the environment to load (default \"l2rpn_case14_sandbox\"")
    parser.add_argument("--logs_dir", required=False, default=None,
                        help="Where to output the training logs (usually tensorboard logs)")
    return parser



# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

import argparse
from l2rpn_baselines.utils.str2bool import str2bool


# TODO add possibilitiy to add new CLI when calling the function
def cli_eval():
    """ some usefull command line arguments (CLI) for the evaluation of the baseline."""
    parser = argparse.ArgumentParser(description="Train baseline DDQN")
    parser.add_argument("--env_name", default="l2rpn_case14_sandbox",
                        help="Name of the l2rpn dataset environment")
    parser.add_argument("--load_path", default=".",
                        help="The path where the trained model has been stored")
    parser.add_argument("--name", default="baseline",
                        help="The name of the trained baseline")
    parser.add_argument("--logs_dir", required=False,
                        default="./logs-eval", type=str,
                        help="Path to output logs directory")
    parser.add_argument("--nb_episode", required=False,
                        default=1, type=int,
                        help="Number of episodes to evaluate")
    parser.add_argument("--nb_process", required=False,
                        default=1, type=int,
                        help="Number of cores to use")
    parser.add_argument("--max_steps", required=False,
                        default=8000, type=int,
                        help="Maximum number of steps per scenario")
    parser.add_argument("--verbose", type=str2bool, nargs='?',
                        const=True, default=False,
                        help="Enable verbose runner mode..")
    parser.add_argument("--save_gif", type=str2bool, nargs='?',
                        const=True, default=False,
                        help="Save the gif as \"epidose.gif\" in the episode path module.")
    return parser



try:
    import grid2op
    import matplotlib
    import numpy as np
    import pandas as pd
    import matplotlib.pyplot as plt
    import warnings
    import os
    from grid2op.MakeEnv import make
    import itertools
    import os.path
except:
    print("You need the following package to train the A3C baseline")
    print(sys.exc_info())
    exit()

def prune_action_space(bus_no):
    # Initialize the env."case5_example",chronics_path=os.path.join("public_data", "chronics_5bus_example")
    if bus_no == "14":
        environment = make()
        num_substations = 14 # 14 for 14 bus
    elif bus_no == "5":
        environment = make("case5_example",chronics_path=os.path.join("public_data", "chronics_5bus_example"))
        num_substations = 5
    else:
        print("Unexpected bus size of the power system. Exiting code in Action_reduced_list.py")
        exit()

    from pandapower.plotting.plotly import simple_plotly
    # simple_plotly(environment.backend._grid)
    # Load ids
    print("\nInjection information:")
    load_to_subid = environment.action_space.load_to_subid
    print ('There are {} loads connected to substations with id: {}'.format(len(load_to_subid), load_to_subid))

    # Generators irds
    gen_to_subid = environment.action_space.gen_to_subid
    print ('There are {} generators, connected to substations with id: {}'.format(len(gen_to_subid), gen_to_subid))

    # Line id sender
    print("\nPowerline information:")
    lines_or_to_subid = environment.action_space.lines_or_to_subid
    lines_ex_to_subid = environment.action_space.lines_ex_to_subid

    print ('There are {} transmissions lines on this grid.'.format(len(lines_or_to_subid)))


    gen_at_sub = []
    load_at_sub = []
    lines_or_at_sub = []
    lines_ex_at_sub = []
    for k in range(num_substations):
        gen_at_sub.append(np.arange(len(gen_to_subid))[gen_to_subid==k])
        load_at_sub.append(np.arange(len(load_to_subid))[load_to_subid==k])
        lines_or_at_sub.append(np.arange(len(lines_or_to_subid))[lines_or_to_subid==k])
        lines_ex_at_sub.append(np.arange(len(lines_ex_to_subid))[lines_ex_to_subid==k])
    1;
    # Num of elements per SE
    print("\nSubstations information:")
    for i, nb_el in enumerate(environment.action_space.subs_info):
        print("On susbtation {} there are {} elements.".format(i, nb_el))

    # adding the change_actions at for 1 substation at a time . Initializing with no action
    gen_action_list =[[]] # list of gens acted
    load_action_list = [[]]
    line_or_action_list = [[]]
    line_ex_action_list = [[]]
    substation_acted = [[]] # substation acted
    gen_action_list_bus_2 = [[]]
    load_action_list_bus_2 = [[]]
    line_or_action_list_bus_2 = [[]]
    line_ex_action_list_bus_2 = [[]]
    do_nothing_act = environment.helper_action_player({})
    obs, reward, done, info = environment.step(do_nothing_act)
    for sub_id in range(num_substations):
        num_gen_at_sub = len(gen_at_sub[sub_id] )
        num_load_at_sub = len(load_at_sub[sub_id] )
        num_line_or_at_sub = len(lines_or_at_sub[sub_id] )
        num_line_ex_at_sub = len(lines_ex_at_sub[sub_id] )
        switching_patterns = ["".join(seq) for seq in itertools.product("01",repeat=environment.action_space.subs_info[sub_id] - 1)]  # reduce by 1 bit due to compliment being the same
        switching_patterns = [[int(sw_i_k) for sw_i_k in '0' + sw_i] for sw_i in switching_patterns]  # adding back the '0' at the beginning as we are fizing this bit
        switching_patterns.pop(0) # deleting first action as it is a no-action
        for sw_action in switching_patterns:
            switching_patterns_split = np.split(np.array(sw_action), np.cumsum([num_gen_at_sub,num_load_at_sub,num_line_or_at_sub,num_line_ex_at_sub]) )
            gen_action_list.append(gen_at_sub[sub_id][switching_patterns_split[0] == 1])
            load_action_list.append(load_at_sub[sub_id][switching_patterns_split[1] == 1])
            line_or_action_list.append(lines_or_at_sub[sub_id][switching_patterns_split[2] == 1])
            line_ex_action_list.append(lines_ex_at_sub[sub_id][switching_patterns_split[3] == 1])
            substation_acted.append(sub_id)
            # gen_action_list_bus_2.append(gen_at_sub[sub_id][switching_patterns_split[0] == 0])
            # load_action_list_bus_2.append(load_at_sub[sub_id][switching_patterns_split[1] == 0])
            # line_or_action_list_bus_2.append(lines_or_at_sub[sub_id][switching_patterns_split[2] == 0])
            # line_ex_action_list_bus_2.append(lines_ex_at_sub[sub_id][switching_patterns_split[3] == 0])
            # do_act = environment.action_space(
            #     {"set_bus": {"generators_id": [(g,1) for g in gen_action_list[-1]]+[(g,2) for g in gen_action_list_bus_2[-1]],
            #                  "loads_id": [(lo,1) for lo in load_action_list[-1]]+[(lo,2) for lo in load_action_list_bus_2[-1]],
            #                  "lines_or_id": [(li,1) for li in line_or_action_list[-1]]+[(li,2) for li in line_or_action_list_bus_2[-1]],
            #                  "lines_ex_id": [(li,1) for li in line_ex_action_list[-1]]+[(li,2) for li in line_ex_action_list_bus_2[-1]]}})
            # print(do_act)
            # obs_sim, reward_sim, is_done_sim, info_sim = obs.simulate(do_act)
            # print(obs_sim)
            # print(is_done_sim)
            # if is_done_sim:
            #     print("----------------------")
            # print(info_sim)
            # h_letters = [(letter, 2) for letter in 'human']
            # print(h_letters)

            # do_nothing_act = environment.action_space({"change_bus":{"generators_id": [0],"loads_id": [1],"lines_or_id":[3],"lines_ex_id":[7]}})
            # obs_sim, reward_sim, is_done_sim, info_sim = obs.simulate(do_nothing_act)
    #
    action_index = 1 # max is 60 for 5 bus
    do_act = environment.action_space(
        {"change_bus": {"generators_id": gen_action_list[action_index], "loads_id": load_action_list[action_index], "lines_or_id": line_or_action_list[action_index], "lines_ex_id": line_ex_action_list[action_index]}})
    print(do_act)
    obs_sim, reward_sim, is_done_sim, info_sim = obs.simulate(do_act)
    k = 1
    # environment.action_space({"change_bus":{"generators_id": [0],"loads_id": [1],"lines_or_id":[3],"lines_ex_id":[7]}})
    np.save("gen_action_list.npy", gen_action_list)
    np.save("load_action_list.npy", load_action_list)
    np.save("line_or_action_list.npy", line_or_action_list)
    np.save("line_ex_action_list.npy", line_ex_action_list)
    return gen_action_list, load_action_list, line_or_action_list, line_ex_action_list

def main_function(bus_no):
    # There are a total of 4 files.
    file_names = ["gen_action_list","load_action_list","line_or_action_list","line_ex_action_list"]
    flag_check = None
    for index_item, item in enumerate(file_names):
        flag_check = os.path.isfile(item+".npy")
        if flag_check == False:
            # run the action pruning code!
            gen_action_list, load_action_list, line_or_action_list, line_ex_action_list = prune_action_space(bus_no)
            return gen_action_list, load_action_list, line_or_action_list, line_ex_action_list
    if flag_check == True:
        # load the data from the saved files.
        gen_action_list = np.load("gen_action_list.npy",allow_pickle=True)
        load_action_list = np.load("load_action_list.npy",allow_pickle=True)
        line_or_action_list = np.load("line_or_action_list.npy",allow_pickle=True)
        line_ex_action_list = np.load("line_ex_action_list.npy",allow_pickle=True)
    return gen_action_list, load_action_list, line_or_action_list, line_ex_action_list



#!/usr/bin/env python3

# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

import os
import tensorflow as tf

from grid2op.MakeEnv import make
from grid2op.Runner import Runner
from grid2op.Reward import *
from grid2op.Action import *

from l2rpn_baselines.utils.save_log_gif import save_log_gif
from l2rpn_baselines.SAC.SAC import SAC, DEFAULT_NAME
from l2rpn_baselines.SAC.SAC_NNParam import SAC_NNParam
from l2rpn_baselines.SAC.SAC_NN import SAC_NN

DEFAULT_LOGS_DIR = "./logs-eval/do-nothing-baseline"
DEFAULT_NB_EPISODE = 1
DEFAULT_NB_PROCESS = 1
DEFAULT_MAX_STEPS = -1


def evaluate(env,
             name=DEFAULT_NAME,
             load_path=None,
             logs_path=DEFAULT_LOGS_DIR,
             nb_episode=DEFAULT_NB_EPISODE,
             nb_process=DEFAULT_NB_PROCESS,
             max_steps=DEFAULT_MAX_STEPS,
             verbose=False,
             save_gif=False):
    """
    How to evaluate the performances of the trained SAC agent.

    Parameters
    ----------
    env: :class:`grid2op.Environment`
        The environment on which you evaluate your agent.

    name: ``str``
        The name of the trained baseline

    load_path: ``str``
        Path where the agent has been stored

    logs_path: ``str``
        Where to write the results of the assessment

    nb_episode: ``str``
        How many episodes to run during the assessment of the performances

    nb_process: ``int``
        On how many process the assessment will be made. (setting this > 1 can lead to some speed ups but can be
        unstable on some plaform)

    max_steps: ``int``
        How many steps at maximum your agent will be assessed

    verbose: ``bool``
        Currently un used

    save_gif: ``bool``
        Whether or not you want to save, as a gif, the performance of your agent. It might cause memory issues (might
        take a lot of ram) and drastically increase computation time.

    Returns
    -------
    agent: :class:`l2rpn_baselines.utils.DeepQAgent`
        The loaded agent that has been evaluated thanks to the runner.

    res: ``list``
        The results of the Runner on which the agent was tested.


    Examples
    -------
    You can evaluate a DeepQSimple this way:

    .. code-block:: python

        from grid2op.Reward import L2RPNSandBoxScore, L2RPNReward
        from l2rpn_baselines.SAC import eval

        # Create dataset env
        env = make("l2rpn_case14_sandbox",
                   reward_class=L2RPNSandBoxScore,
                   other_rewards={
                       "reward": L2RPNReward
                   })

        # Call evaluation interface
        evaluate(env,
                 name="MyAwesomeAgent",
                 load_path="/WHERE/I/SAVED/THE/MODEL",
                 logs_path=None,
                 nb_episode=10,
                 nb_process=1,
                 max_steps=-1,
                 verbose=False,
                 save_gif=False)
    """

    # Limit gpu usage
    physical_devices = tf.config.list_physical_devices('GPU')
    if len(physical_devices):
        tf.config.experimental.set_memory_growth(physical_devices[0], True)

    runner_params = env.get_params_for_runner()
    runner_params["verbose"] = verbose

    if load_path is None:
        raise RuntimeError("Cannot evaluate a model if there is nothing to be loaded.")
    path_model, path_target_model = SAC_NN.get_path_model(load_path, name)
    nn_archi = SAC_NNParam.from_json(os.path.join(path_model, "nn_architecture.json"))

    # Run
    # Create agent
    agent = SAC(action_space=env.action_space,
                name=name,
                store_action=nb_process == 1,
                nn_archi=nn_archi)

    # Load weights from file
    agent.load(load_path)
    agent.init_obs_extraction(env)

    # Print model summary
    stringlist = []
    agent.deep_q.model_value.summary(print_fn=lambda x: stringlist.append(x))
    short_model_summary = "\n".join(stringlist)

    if verbose:
        print("Value model: {}".format(short_model_summary))

    # Build runner
    runner = Runner(**runner_params,
                    agentClass=None,
                    agentInstance=agent)

    # Run
    os.makedirs(logs_path, exist_ok=True)
    res = runner.run(path_save=logs_path,
                     nb_episode=nb_episode,
                     nb_process=nb_process,
                     max_iter=max_steps,
                     pbar=verbose)

    # Print summary

    if verbose:
        print("Evaluation summary:")
        for _, chron_name, cum_reward, nb_time_step, max_ts in res:
            msg_tmp = "chronics at: {}".format(chron_name)
            msg_tmp += "\ttotal score: {:.6f}".format(cum_reward)
            msg_tmp += "\ttime steps: {:.0f}/{:.0f}".format(nb_time_step, max_ts)
            print(msg_tmp)

        if len(agent.dict_action):
            # I output some of the actions played
            print("The agent played {} different action".format(len(agent.dict_action)))
            for id_, (nb, act, types) in agent.dict_action.items():
                print("Action with ID {} was played {} times".format(id_, nb))
                print("{}".format(act))
                print("-----------")

    if save_gif:
        if verbose:
            print("Saving the gif of the episodes")
        save_log_gif(logs_path, res)

    return agent, res


if __name__ == "__main__":
    from grid2op.Reward import L2RPNSandBoxScore, L2RPNReward
    from l2rpn_baselines.utils import cli_eval

    # Parse command line
    args = cli_eval().parse_args()

    # Create dataset env
    env = make(args.env_name,
               reward_class=L2RPNSandBoxScore,
               other_rewards={
                   "reward": L2RPNReward
               })

    # Call evaluation interface
    evaluate(env,
             name=args.name,
             load_path=os.path.abspath(args.load_path),
             logs_path=args.logs_dir,
             nb_episode=args.nb_episode,
             nb_process=args.nb_process,
             max_steps=args.max_steps,
             verbose=args.verbose,
             save_gif=args.save_gif)



__all__ = [
    "DeepQSimple",
    "evaluate",
    "train",
    "DeepQ_NNParam"
]

from l2rpn_baselines.DeepQSimple.DeepQSimple import DeepQSimple
from l2rpn_baselines.DeepQSimple.evaluate import evaluate
from l2rpn_baselines.DeepQSimple.train import train
from l2rpn_baselines.DeepQSimple.DeepQ_NNParam import DeepQ_NNParam



__all__ = [
    "ActorCritic_Agent",
    "evaluate",
    "train",
    "Action_reduced_list",
    "Runner",
    "run_grid2viz",
    "user_environment_make"
]

from l2rpn_baselines.Multithreading_agent.ActorCritic_Agent import ActorCritic_Agent
from l2rpn_baselines.Multithreading_agent.evaluate import evaluate
from l2rpn_baselines.Multithreading_agent.train import train
from l2rpn_baselines.Multithreading_agent.Action_reduced_list import Action_reduced_list
from l2rpn_baselines.Multithreading_agent.Runner import Runner
from l2rpn_baselines.Multithreading_agent.run_grid2viz import run_grid2viz
from l2rpn_baselines.Multithreading_agent.user_environment_make import user_environment_make



# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

import warnings
from grid2op.Environment import Environment
try:
    from grid2op.Environment import MultiEnvironment
except ImportError:
    # name will be change as of grid2op >= 1.0.0
    try:
        from grid2op.Environment import MultiEnvMultiProcess as MultiEnvironment
    except ImportError as exc:
        raise exc


def make_multi_env(env_init, nb_env):
    """
    This function creates a multi environment compatible with what is expected in the baselines. In particular, it
    adds the observation_space, the action_space and the reward_range attribute.

    The way this function works is explained in the getting_started of grid2op.

    Attributes
    -----------
    env_init: :class:`grid2op.Environment.Environment`
        The environment to duplicates
    nb_env: ``int``
        The number of environment on with which you want to interact at the same time

    Returns
    -------
    res: :class:`grid2op.Environment.MultiEnvironment` or :class:`grid2op.Environment.Environment`
        A copy of the initial environment (if nb_env = 1) or a MultiEnvironment based on the initial environment
        if nb_env >= 2.

    """
    res = None
    nb_env = int(nb_env)

    if nb_env <= 0:
        raise RuntimeError("Impossible to create a negative number of environments")

    if nb_env == 1:
        warnings.warn("You asked to create 1 environment. We didn't use the MultiEnvironment for that. We instead "
                      "created a copy of your initial environment.")
        res = Environment(**env_init.get_kwargs())
    else:
        res = MultiEnvironment(nb_env, env_init)
        res.observation_space = env_init.observation_space
        res.action_space = env_init.action_space
        res.reward_range = env_init.reward_range
    return res


# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

import argparse


def str2bool(v):
    """INTERNAL DO NOT USE"""
    if isinstance(v, bool):
       return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


"""
https://github.com/MorvanZhou/Reinforcement-learning-with-tensorflow/blob/master/contents/5.2_Prioritized_Replay_DQN/RL_brain.py
"""

import numpy as np

class SumTree(object):
  """
  This SumTree code is a modified version and the original code is from:
  https://github.com/jaara/AI-blog/blob/master/SumTree.py
  Story data with its priority in the tree.
  """
  data_pointer = 0

  def __init__(self, capacity):
    self.capacity = capacity  # for all priority values
    self.tree = np.zeros(2 * capacity - 1)
    # [--------------Parent nodes-------------][-------leaves to recode priority-------]
    #             size: capacity - 1                       size: capacity
    self.data = np.zeros(capacity, dtype=object)  # for all transitions
    # [--------------data frame-------------]
    #             size: capacity

  def add(self, p, data):
    tree_idx = self.data_pointer + self.capacity - 1
    self.data[self.data_pointer] = data  # update data_frame
    self.update(tree_idx, p)  # update tree_frame

    self.data_pointer += 1
    if self.data_pointer >= self.capacity:  # replace when exceed the capacity
      self.data_pointer = 0

  def update(self, tree_idx, p):
    change = p - self.tree[tree_idx]
    self.tree[tree_idx] = p
    # then propagate the change through tree
    while tree_idx != 0:    # this method is faster than the recursive loop in the reference code
      tree_idx = (tree_idx - 1) // 2
      self.tree[tree_idx] += change

  def get_leaf(self, v):
    """
    Tree structure and array storage:
    Array type for storing:
    [0,1,2,3,4,5,6]
    """
    parent_idx = 0
    while True:     # the while loop is faster than the method in the reference code
      cl_idx = 2 * parent_idx + 1         # this leaf's left and right kids
      cr_idx = cl_idx + 1
      if cl_idx >= len(self.tree):        # reach bottom, end search
        leaf_idx = parent_idx
        break
      else:       # downward search, always search for a higher priority node
        if v <= self.tree[cl_idx]:
          parent_idx = cl_idx
        else:
          v -= self.tree[cl_idx]
          parent_idx = cr_idx

    data_idx = leaf_idx - self.capacity + 1
    return leaf_idx, self.tree[leaf_idx], self.data[data_idx]

  @property
  def total_p(self):
    return self.tree[0]  # the root


class Memory(object):  # stored as ( s, a, r, s_ ) in SumTree
  """
  This Memory class is modified based on the original code from:
  https://github.com/jaara/AI-blog/blob/master/Seaquest-DDQN-PER.py
  """
  epsilon = 0.01  # small amount to avoid zero priority
  beta_increment_per_sampling = 0.001
  abs_err_upper = 1.  # clipped abs error

  def __init__(self, capacity, alpha=0.6, beta=0.4):
    self.capacity = capacity
    self.tree = SumTree(capacity)
    self.alpha = alpha
    self.beta = beta

  def store(self, transition):
    max_p = np.max(self.tree.tree[-self.tree.capacity:])
    if max_p == 0:
      max_p = self.abs_err_upper
    self.tree.add(max_p, transition)   # set the max p for new p

  def sample(self, n):
    b_idx, b_memory, ISWeights = np.empty((n,), dtype=np.int32), [], []
    pri_seg = self.tree.total_p / n       # priority segment
    self.beta = np.min(
        [1., self.beta + self.beta_increment_per_sampling])  # max = 1

    # for later calculate ISweight
    # min_prob = np.min(self.tree.tree[-self.tree.capacity:]) / self.tree.total_p
    for i in range(n):
      
      a, b = pri_seg * i, pri_seg * (i + 1)
      try: v = np.random.uniform(a, b)
      except:v = np.random.uniform(b, a)
      idx, p, data = self.tree.get_leaf(v)
      prob = p / self.tree.total_p
      ISWeights.append(np.power(self.capacity * prob, -self.beta))
      b_idx[i] = idx
      b_memory.append(data)
    max_weight = max(ISWeights)
    ISWeights = [x / max_weight for x in ISWeights]
    return b_idx, b_memory, ISWeights

  def batch_update(self, tree_idx, abs_errors):
    abs_errors += self.epsilon  # convert to abs and avoid 0
    clipped_errors = np.minimum(abs_errors, self.abs_err_upper)
    ps = np.power(clipped_errors, self.alpha)
    for ti, p in zip(tree_idx, ps):
      self.tree.update(ti, p)


if __name__ == "__main__":
  exp = Memory(32)
  for i in range(10):
    exp.store(np.array([[1, 2], 2, 3, [2, 3], 5]))

  b_idx, b_memory, ISWeights = exp.sample(2)
  print(b_idx, b_memory, ISWeights)



# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.

import setuptools
from setuptools import setup
__version__ = "0.4.1"


pkgs = {
    "required": [
        "grid2op[optional]>=0.9.1.post1",
        "tensorflow>=2.2.0",
        "Keras>=2.3.1",
        "torch>=1.4.0",
        "statsmodels>=0.11.1",
        "scikit-learn>=0.22.2.post1",
        "gym>=0.17.1",
        "scipy>=1.4.1",
    ],
    "extras": {
        "docs": [
            "numpydoc>=0.9.2",
            "sphinx>=2.4.4",
            "sphinx-rtd-theme>=0.4.3",
            "sphinxcontrib-trio>=1.1.0",
            "autodocsumm>=0.1.13"
        ],
        "challenge": ["grid2op[challenge]>=0.9.1.post1"]
    }
}


setup(name='l2rpn_baselines',
      version=__version__,
      description='L2RPN Baselines a repository to host ' \
      'baselines for l2rpn competitions.',
      long_description='This repository aims at facilitating ' \
      'the use of state of the art algorithm in coming from the ' \
      'reinforcement learning community or the power system ' \
      'community in the l2rpn competitions. It  also provides ' \
      'some usefull function to make life or participants to the ' \
      'l2rpn competitions easier.',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
          "Intended Audience :: Developers",
          "Intended Audience :: Education",
          "Intended Audience :: Science/Research",
          "Natural Language :: English"
      ],
      keywords='ML powergrid optmization RL power-systems',
      author='Benjamin DONNOT',
      author_email='benjamin.donnot@rte-france.com',
      url="https://github.com/BDonnot/L2RPN_Baselines",
      license='MPL',
      packages=setuptools.find_packages(),
      include_package_data=True,
      install_requires=pkgs["required"],
      extras_require=pkgs["extras"],
      zip_safe=False,
      entry_points={}
)



try:
    import grid2op
    import matplotlib
    import numpy as np
    import pandas as pd
    import matplotlib.pyplot as plt
    import warnings
    import os
    from grid2op.MakeEnv import make
    import itertools
    import os.path
except ImportError as exc_:
    raise ImportError("AsynchronousActorCritic baseline impossible to load the required dependencies. The error was: \n {}".format(exc_))


def prune_action_space(bus_no):
    # Initialize the env."case5_example",chronics_path=os.path.join("public_data", "chronics_5bus_example")
    if bus_no == "14":
        environment = make()
        num_substations = 14 # 14 for 14 bus
    elif bus_no == "5":
        environment = make("case5_example",chronics_path=os.path.join("public_data", "chronics_5bus_example"))
        num_substations = 5
    else:
        print("Unexpected bus size of the power system. Exiting code in Action_reduced_list.py")
        exit()

    from pandapower.plotting.plotly import simple_plotly
    # simple_plotly(environment.backend._grid)
    # Load ids
    print("\nInjection information:")
    load_to_subid = environment.action_space.load_to_subid
    print ('There are {} loads connected to substations with id: {}'.format(len(load_to_subid), load_to_subid))

    # Generators irds
    gen_to_subid = environment.action_space.gen_to_subid
    print ('There are {} generators, connected to substations with id: {}'.format(len(gen_to_subid), gen_to_subid))

    # Line id sender
    print("\nPowerline information:")
    lines_or_to_subid = environment.action_space.lines_or_to_subid
    lines_ex_to_subid = environment.action_space.lines_ex_to_subid

    print ('There are {} transmissions lines on this grid.'.format(len(lines_or_to_subid)))


    gen_at_sub = []
    load_at_sub = []
    lines_or_at_sub = []
    lines_ex_at_sub = []
    for k in range(num_substations):
        gen_at_sub.append(np.arange(len(gen_to_subid))[gen_to_subid==k])
        load_at_sub.append(np.arange(len(load_to_subid))[load_to_subid==k])
        lines_or_at_sub.append(np.arange(len(lines_or_to_subid))[lines_or_to_subid==k])
        lines_ex_at_sub.append(np.arange(len(lines_ex_to_subid))[lines_ex_to_subid==k])
    1;
    # Num of elements per SE
    print("\nSubstations information:")
    for i, nb_el in enumerate(environment.action_space.subs_info):
        print("On susbtation {} there are {} elements.".format(i, nb_el))

    # adding the change_actions at for 1 substation at a time . Initializing with no action
    gen_action_list =[[]] # list of gens acted
    load_action_list = [[]]
    line_or_action_list = [[]]
    line_ex_action_list = [[]]
    substation_acted = [[]] # substation acted
    gen_action_list_bus_2 = [[]]
    load_action_list_bus_2 = [[]]
    line_or_action_list_bus_2 = [[]]
    line_ex_action_list_bus_2 = [[]]
    do_nothing_act = environment.helper_action_player({})
    obs, reward, done, info = environment.step(do_nothing_act)
    for sub_id in range(num_substations):
        num_gen_at_sub = len(gen_at_sub[sub_id] )
        num_load_at_sub = len(load_at_sub[sub_id] )
        num_line_or_at_sub = len(lines_or_at_sub[sub_id] )
        num_line_ex_at_sub = len(lines_ex_at_sub[sub_id] )
        switching_patterns = ["".join(seq) for seq in itertools.product("01",repeat=environment.action_space.subs_info[sub_id] - 1)]  # reduce by 1 bit due to compliment being the same
        switching_patterns = [[int(sw_i_k) for sw_i_k in '0' + sw_i] for sw_i in switching_patterns]  # adding back the '0' at the beginning as we are fizing this bit
        switching_patterns.pop(0) # deleting first action as it is a no-action
        for sw_action in switching_patterns:
            switching_patterns_split = np.split(np.array(sw_action), np.cumsum([num_gen_at_sub,num_load_at_sub,num_line_or_at_sub,num_line_ex_at_sub]) )
            gen_action_list.append(gen_at_sub[sub_id][switching_patterns_split[0] == 1])
            load_action_list.append(load_at_sub[sub_id][switching_patterns_split[1] == 1])
            line_or_action_list.append(lines_or_at_sub[sub_id][switching_patterns_split[2] == 1])
            line_ex_action_list.append(lines_ex_at_sub[sub_id][switching_patterns_split[3] == 1])
            substation_acted.append(sub_id)
            # gen_action_list_bus_2.append(gen_at_sub[sub_id][switching_patterns_split[0] == 0])
            # load_action_list_bus_2.append(load_at_sub[sub_id][switching_patterns_split[1] == 0])
            # line_or_action_list_bus_2.append(lines_or_at_sub[sub_id][switching_patterns_split[2] == 0])
            # line_ex_action_list_bus_2.append(lines_ex_at_sub[sub_id][switching_patterns_split[3] == 0])
            # do_act = environment.action_space(
            #     {"set_bus": {"generators_id": [(g,1) for g in gen_action_list[-1]]+[(g,2) for g in gen_action_list_bus_2[-1]],
            #                  "loads_id": [(lo,1) for lo in load_action_list[-1]]+[(lo,2) for lo in load_action_list_bus_2[-1]],
            #                  "lines_or_id": [(li,1) for li in line_or_action_list[-1]]+[(li,2) for li in line_or_action_list_bus_2[-1]],
            #                  "lines_ex_id": [(li,1) for li in line_ex_action_list[-1]]+[(li,2) for li in line_ex_action_list_bus_2[-1]]}})
            # print(do_act)
            # obs_sim, reward_sim, is_done_sim, info_sim = obs.simulate(do_act)
            # print(obs_sim)
            # print(is_done_sim)
            # if is_done_sim:
            #     print("----------------------")
            # print(info_sim)
            # h_letters = [(letter, 2) for letter in 'human']
            # print(h_letters)

            # do_nothing_act = environment.action_space({"change_bus":{"generators_id": [0],"loads_id": [1],"lines_or_id":[3],"lines_ex_id":[7]}})
            # obs_sim, reward_sim, is_done_sim, info_sim = obs.simulate(do_nothing_act)
    #
    action_index = 1 # max is 60 for 5 bus
    do_act = environment.action_space(
        {"change_bus": {"generators_id": gen_action_list[action_index], "loads_id": load_action_list[action_index], "lines_or_id": line_or_action_list[action_index], "lines_ex_id": line_ex_action_list[action_index]}})
    print(do_act)
    obs_sim, reward_sim, is_done_sim, info_sim = obs.simulate(do_act)
    k = 1
    # environment.action_space({"change_bus":{"generators_id": [0],"loads_id": [1],"lines_or_id":[3],"lines_ex_id":[7]}})
    np.save("gen_action_list.npy", gen_action_list)
    np.save("load_action_list.npy", load_action_list)
    np.save("line_or_action_list.npy", line_or_action_list)
    np.save("line_ex_action_list.npy", line_ex_action_list)
    return gen_action_list, load_action_list, line_or_action_list, line_ex_action_list

def main_function(bus_no):
    # There are a total of 4 files.
    file_names = ["gen_action_list","load_action_list","line_or_action_list","line_ex_action_list"]
    flag_check = None
    for index_item, item in enumerate(file_names):
        flag_check = os.path.isfile(item+".npy")
        if flag_check == False:
            # run the action pruning code!
            gen_action_list, load_action_list, line_or_action_list, line_ex_action_list = prune_action_space(bus_no)
            return gen_action_list, load_action_list, line_or_action_list, line_ex_action_list
    if flag_check == True:
        # load the data from the saved files.
        gen_action_list = np.load("gen_action_list.npy",allow_pickle=True)
        load_action_list = np.load("load_action_list.npy",allow_pickle=True)
        line_or_action_list = np.load("line_or_action_list.npy",allow_pickle=True)
        line_ex_action_list = np.load("line_ex_action_list.npy",allow_pickle=True)
    return gen_action_list, load_action_list, line_or_action_list, line_ex_action_list



from grid2op import make
from grid2op.Parameters import Parameters
from grid2op.Reward import L2RPNReward, CombinedReward, CloseToOverflowReward, GameplayReward

def set_environement(start_id,env_name,profiles_chronics):
    param = Parameters()
    param.NO_OVERFLOW_DISCONNECTION = True

    env = make(env_name,chronics_path= profiles_chronics, reward_class=CombinedReward,param=param)
    # Register custom reward for training
    cr = env.reward_helper.template_reward
    cr.addReward("overflow", CloseToOverflowReward(), 50.0)
    cr.addReward("game", GameplayReward(), 100.0)
    cr.initialize(env)

    # Debug prints
    print("Debug prints --->:")
    print("Chronics location that being used:", env.chronics_handler.path)
    print("Grid location being used:", env.init_grid_path)
    print("Reward class that is being used:", env.rewardClass)
    print("Action type class being used:", env.actionClass)
    print("Observation type class being used:", env.observationClass)
    print("Backend CSV file key names:", env.names_chronics_to_backend)
    print("Legal action class being used:", env.legalActClass)
    print("Voltage controller class being used:", env.voltagecontrolerClass)

    if start_id != None:
        env.chronics_handler.tell_id(start_id)
        print("Thread number:",start_id,", ID of chronic current folder:",env.chronics_handler.real_data.id_chron_folder_current)
    return env



__all__ = [
    "Geirina",
    "evaluate",
    "train"
]

from l2rpn_baselines.Geirina.Geirina import Geirina
from l2rpn_baselines.Geirina.evaluate import evaluate
from l2rpn_baselines.Geirina.train import train

"""
In the __init__ file, it is expected to export 3 classes with names that depends on the name you gave to your baseline.
For example, say you chose to write a baseline with the awesome name "XXX" (what an imagination!) you should export
in this __init__.py file:

- `XXX` [**mandatory**] contains the definition of your baseline. It must follow the directives
   given in "Template.py"
- `evaluate` [**mandatory**] contains the script to evaluate the performance of this baseline. It must
  follow the directive in "evaluate.py"
- `train` [**optional**] contains the script to train your baseline. If provided, it must follow
  the directives given in "train.py"
  
See the import above for an example on how to export your scripts properly.
"""
